"""
Workflow class provides all the workflow data
required by MS Transferor
"""
from __future__ import division, print_function

from builtins import range, object
from future.utils import viewitems, viewvalues, listvalues

from copy import copy, deepcopy
from WMCore.DataStructs.LumiList import LumiList
from WMCore.MicroService.Tools.Common import getMSLogger, isRelVal
from WMCore.Services.Rucio.RucioUtils import GROUPING_DSET, GROUPING_ALL, NUM_COPIES_DEFAULT


class Workflow(object):
    """
    Class to represent a workflow and some helpers to access
    its information within MS
    """

    def __init__(self, reqName, reqData, logger=None, verbose=False):
        self.reqName = reqName
        self.data = reqData
        # TODO: this replace can be removed in one year from now, thus March 2022
        self.data['DbsUrl'] = self.data['DbsUrl'].replace("cmsweb.cern.ch", "cmsweb-prod.cern.ch")
        # stripping any end slashes, which no longer work in the Go-based server
        self.data['DbsUrl'] = self.data['DbsUrl'].rstrip("/")

        self.logger = getMSLogger(verbose, logger)

        self.inputDataset = ""
        self.parentDataset = ""
        self.pileupDatasets = set()
        self.pileupRSEList = set()

        self.campaigns = set()
        self.dataCampaignMap = []
        # these blocks structure will be key'ed by the block name and value'd by the block size
        self.primaryBlocks = {}
        self.parentBlocks = {}
        # sort of duplicate info, but we need to have a way to link input to parent block(s)
        self.childToParentBlocks = {}
        # pileup don't need to get resolved into blocks, store only their total size and location
        self.secondarySummaries = {}

        self.setDataCampaignMap()
        self.setInputData()

    def __str__(self):
        """
        Write out useful information for this object
        :return:
        """
        res = {'reqName': self.reqName, 'inputDataset': self.inputDataset,
               'parentDataset': self.parentDataset, 'pileupDatasets': self.pileupDatasets,
               'campaigns': self.campaigns, 'dataCampaignMap': self.dataCampaignMap}
        return str(res)

    def getName(self):
        """
        Get this request name
        """
        return self.reqName

    def getDbsUrl(self):
        """
        Get the DbsUrl defined in this request
        """
        return self.data['DbsUrl']

    def getReqType(self):
        """
        Return the request type for this workflow
        """
        return self.data['RequestType']

    def getReqParam(self, param):
        """
        Return a top level parameter for this request.
        Read: parameters internal to Task/Steps are not looked up.
        """
        if param not in self.data:
            self.logger.warning("Request parameter '%s' not found in the workflow %s", param, self.getName())
        return self.data.get(param)

    def getSitelist(self):
        """
        Get the SiteWhitelist minus the black list for this request
        """
        return sorted(list(set(self.data['SiteWhitelist']) - set(self.data['SiteBlacklist'])))

    def setPURSElist(self, rseSet):
        """
        Hook/hack to make sure that multiple pileup datasets are placed
        in the same storage (unless secondaryAAA is enabled).
        It will be used to place the primary/parent blocks under the same
        location too (in case primaryAAA is enabled)
        Set this location only once, any further updates should not work!
        :param rseSet: a set of RSE names
        """
        if isinstance(rseSet, set):
            self.pileupRSEList = rseSet
        elif isinstance(rseSet, list):
            self.pileupRSEList = set(rseSet)
        else:
            # assume it's a string
            self.pileupRSEList = {rseSet}

    def getPURSElist(self):
        """
        Retrieve the final list of RSEs where ALL the data needs to be placed
        """
        return self.pileupRSEList

    def getRunWhitelist(self):
        """
        Get the RunWhitelist for this request
        """
        res = set(self._getValue('RunWhitelist', []))
        return sorted(list(res))

    def getRunBlacklist(self):
        """
        Get the RunBlacklist for this request
        """
        res = set(self._getValue('RunBlacklist', []))
        return sorted(list(res))

    def getLumilist(self):
        """
        Get the LumiList parameter and return a LumiList object,
        in case the LumiList is not empty.
        """
        lumiDict = self._getValue('LumiList', {})
        if not lumiDict:
            return {}
        return LumiList(compactList=lumiDict)

    def setDataCampaignMap(self):
        """
        Set the association between input data, data type and campaign name
        in the format of:
        {"type": "type of input data", "name": "dataset name", "campaign": "campaign name"}
        Also set a flat set of campaign names
        """
        if self.data['RequestType'] in ("StoreResults", "Resubmission"):
            self.logger.info("Request type %s does not support input data placement", self.data['RequestType'])
            self.dataCampaignMap = []
            return

        inputMap = {"InputDataset": "primary", "MCPileup": "secondary", "DataPileup": "secondary"}

        if "TaskChain" in self.data or "StepChain" in self.data:
            innerDicts = []
            for i in range(1, self.data.get("TaskChain", self.data.get("StepChain")) + 1):
                innerDicts.append(self.data.get("Task%d" % i, self.data.get("Step%d" % i)))
        else:
            # ReReco and DQMHarvesting
            innerDicts = [self.data]

        data = {}
        for item in innerDicts:
            for key in inputMap:
                if key in item and item[key]:
                    # use top level campaign if not available in the inner dict
                    data[item[key]] = dict(type=inputMap[key], name=item[key],
                                           campaign=item.get('Campaign', self.data.get('Campaign')))
            # also create a flat list of campaign names
            if "Campaign" in item and item["Campaign"]:
                self.campaigns.add(item["Campaign"])

        self.dataCampaignMap = listvalues(data)

    def getDataCampaignMap(self):
        """
        Retrieve map of campaign, dataset, dataset type
        :return: list of dictionaries
        """
        return self.dataCampaignMap

    def getCampaigns(self):
        """
        Get a set of campaign names used within this request
        """
        return self.campaigns

    def setInputData(self):
        """
        Parse the request data and set the primary and secondary datasets
        """
        for item in self.getDataCampaignMap():
            if item["type"] == "primary":
                self.inputDataset = item["name"]
            elif item["type"] == "secondary":
                self.pileupDatasets.add(item["name"])

    def getInputDataset(self):
        """
        Get this request's input dataset name
        """
        return self.inputDataset

    def getPileupDatasets(self):
        """
        Get this request's secondary dataset names
        """
        return self.pileupDatasets

    def hasParents(self):
        """
        Check whether the request has IncludeParents
        :return:
        """
        if not self.inputDataset:
            return False
        return self._getValue("IncludeParents", False)

    def setParentDataset(self, parent):
        """
        Set the parent dataset name and update the data/campaign map
        :param parent: string corresponding to the parent dataset name
        """
        self.parentDataset = parent
        self._updateDataCampaignMap(parent)

    def getParentDataset(self):
        """
        Return the parent dataset name
        :return: parent dataset name string
        """
        return self.parentDataset

    def getBlockWhitelist(self):
        """
        Fetch the BlockWhitelist for this workflow
        :return: a list data type with the blocks white listed
        """
        return self._getValue("BlockWhitelist", [])

    def getBlockBlacklist(self):
        """
        Fetch the BlockBlacklist for this workflow
        :return: a list data type with the blocks white listed
        """
        return self._getValue("BlockBlacklist", [])

    def setPrimaryBlocks(self, blocksDict):
        """
        Sets a list of primary input blocks taking into consideration
        the BlockWhitelist and BlockBlacklist
        Data is in the form of:
        {"block_name": {"blockSize": 1234,
                        "locations": [list of locations]}}
        """
        blockWhite = self.getBlockWhitelist()
        if blockWhite:
            for block in blockWhite:
                if block in blocksDict:
                    self.primaryBlocks[block] = copy(blocksDict[block])
        else:
            self.primaryBlocks = deepcopy(blocksDict)

        for block in self.getBlockBlacklist():
            self.primaryBlocks.pop(block, None)

    def getPrimaryBlocks(self):
        """
        Retrieve list of input primary blocks
        """
        return self.primaryBlocks

    def setSecondarySummary(self, dsetName, dsetSize, locations=None):
        """
        Create a summary of the pileup dataset, with its total data size
        and locations where the whole dataset is subscribed and available
        :param dsetName: string with the secondary dataset name
        :param dsetSize: integer with the secondary dataset size
        :param locations: locations hosting this dataset in full (and subscribed)
        Data is in the form of:
        {"dataset_name": {"dsetSize": 1234,
                          "locations": [list of locations]}}
        """
        self.secondarySummaries.setdefault(dsetName, {})
        self.secondarySummaries[dsetName]['dsetSize'] = dsetSize
        self.secondarySummaries[dsetName]['locations'] = locations or []

    def getSecondarySummary(self):
        """
        Retrieve list of input secondary datasets and their sizes
        """
        return self.secondarySummaries

    def setParentBlocks(self, blocksDict):
        """
        Sets a list of parent input blocks and their size.
        NOTE: this list is solely based on the parent dataset, without
        considering what are the actual input primary blocks
        {"block_name": {"blockSize": 1234,
                        "locations": [list of locations]}}
        """
        self.parentBlocks = blocksDict

    def getParentBlocks(self):
        """
        Retrieve list of input parent blocks
        """
        return self.parentBlocks

    def setChildToParentBlocks(self, blocksDict):
        """
        Sets a relationship between input primary block and its parent block(s)
        This method guarantees that only parent blocks with valid replicas are
        kept around (and later transferred)
        The child block list is also final, meaning all the run/block/lumi lists
        have already been applied.
        :param blocksDict: dict key'ed by the primary block, with a list of parent blocks
        """
        # flat list with the final parent blocks
        parentBlocks = set()
        # remove parent blocks without any valid replica (only invalid files)
        for child, parents in viewitems(blocksDict):
            for parent in list(parents):
                if parent not in self.getParentBlocks():
                    # then drop this block
                    parents.remove(parent)
            self.childToParentBlocks[child] = blocksDict[child]
            parentBlocks = parentBlocks | set(blocksDict[child])

        # Now remove any parent block that don't need to be transferred
        for block in list(self.getParentBlocks()):
            if block not in parentBlocks:
                self.parentBlocks.pop(block, None)

    def getChildToParentBlocks(self):
        """
        Returns a dictionary of blocks and its correspondent list of parents
        """
        return self.childToParentBlocks

    def _getValue(self, keyName, defaultValue=None):
        """
        Provide a property/keyName, return its valid value if any
        :param property: dictionary key name
        :param defaultValue: default value in case the key is not found
        :return: valid value (non-empty/false) or None
        """
        resp = defaultValue
        if keyName in self.data and self.data[keyName]:
            resp = self.data[keyName]
        elif keyName in self.data.get("Task1", {}):
            resp = self.data["Task1"][keyName]
        elif keyName in self.data.get("Step1", {}):
            resp = self.data["Step1"][keyName]
        return resp

    def _updateDataCampaignMap(self, parentName):
        """
        Update the data/campaign map with a newly discovered parent dataset
        :param parentName: the parent dataset name
        """
        for item in self.getDataCampaignMap():
            if item["type"] == "primary":
                # this is the campaign name we're looking for
                newItem = dict(type="parent", name=parentName, campaign=item["campaign"])
                break
        self.dataCampaignMap.append(newItem)

    def getWorkflowGroup(self):
        """
        Defines a workflow according to its group/activity, such as:
          *) release validation workflows
          *) standard central production workflows
        :return: a string with the workflow class: relval, processing
        """
        if isRelVal(self.data):
            return "relval"
        return "production"

    def getOpenRunningTimeout(self):
        """
        Retrieve the OpenRunningTimeout parameter for this workflow
        :return: an integer with the amount of secs
        """
        return self.data.get("OpenRunningTimeout", 0)

    def getInputData(self):
        """
        Returns all the primary and parent data that has to be locked
        and transferred with Rucio
        :return: a list of unique block names and an integer
                 with their total size
        """
        blockList = list(self.getPrimaryBlocks())
        totalBlockSize = sum([blockInfo['blockSize'] for blockInfo in self.getPrimaryBlocks().values()])

        # if it has parent blocks, add all of them as well
        if self.getParentDataset():
            blockList.extend(list(self.getParentBlocks()))
            totalBlockSize += sum([blockInfo['blockSize'] for blockInfo in viewvalues(self.getParentBlocks())])
        return blockList, totalBlockSize

    def getRucioGrouping(self):
        """
        Returns the rucio rule grouping to be defined for a primary
        and/or parent input data placement, where:
            * ALL: all CMS blocks are placed under the same RSE
            * DATASET: CMS blocks can be scattered in multiple RSEs

        NOTE that this does not apply to secondary data placement,
        which is always "ALL" (whole container in the same RSE).

        :return: a string with the required DID grouping
        """
        if self.getParentDataset():
            return GROUPING_ALL
        return GROUPING_DSET

    def getReplicaCopies(self):
        """
        Returns the number of replica copies to be defined in
        a given rucio rule. Standard/default value is 1.

        :return: an integer with the number of copies
        """
        return NUM_COPIES_DEFAULT



"""
provide the map between configuration of web tool and
DBFactory parameters
"""
from WMCore.Database.ConfigDBMap import ConfigDBMapInterface

class ConfigDBMap(ConfigDBMapInterface):

    def __init__(self, config):
        assert hasattr(config, 'database'), "No database configured"
        if hasattr(config.database, 'instances'):
            # The application has multiple instances
            database = getattr(config.database.instances, config.instance)
            self.configure_my_db(database)
        else:
            self.configure_my_db(config.database)

    def configure_my_db(self, database):
        assert hasattr(database, 'connectUrl'), "No database url configured"
        self.dbUrl = database.connectUrl
        self.option = {}
        if hasattr(database, "socket"):
            self.option["unix_socket"] = database.socket
        if hasattr(database, "engineParameters"):
            self.option['engine_parameters'] = database.engineParameters

    def getDBUrl(self):
        return self.dbUrl

    def getOption(self):
        return self.option



#!/usr/bin/python

"""
_Create_

Class for creating Oracle specific schema for persistent messages.

"""





import threading

from WMCore.Database.DBCreator import DBCreator

class Create(DBCreator):
    """
    _Create_

    Class for creating Oracle specific schema for persistent messages.
    """

    sequence_tables = []
    sequence_tables.append('tp_threadpool_seq')
    sequence_tables.append('tp_buffer_in_seq')
    sequence_tables.append('tp_buffer_out_seq')

    trigger_tables = []
    trigger_tables.append('tp_threadpool_trg')
    trigger_tables.append('tp_buffer_in_trg')
    trigger_tables.append('tp_buffer_out_trg')

    def __init__(self, logger = None, dbi = None, params = None):
        myThread = threading.currentThread()
        DBCreator.__init__(self, myThread.logger, myThread.dbi)
        self.create = {}
        self.constraints = {}

        tablespaceTable = ""
        tablespaceIndex = ""

        if params:
            if "tablespace_table" in params:
                tablespaceTable = "TABLESPACE %s" % params["tablespace_table"]
            if "tablespace_index" in params:
                tablespaceIndex = "USING INDEX TABLESPACE %s" % params["tablespace_index"]

        self.create["tp_threadpool"] = \
          """CREATE TABLE tp_threadpool (
               id             NUMBER(11)    NOT NULL ENABLE,
               event          VARCHAR2(255) NOT NULL ENABLE,
               component      VARCHAR2(255) NOT NULL ENABLE,
               payload        CLOB          NOT NULL ENABLE,
               thread_pool_id VARCHAR2(255) NOT NULL ENABLE,
               state          VARCHAR(20)   DEFAULT 'queued'  NOT NULL ENABLE
               ) %s""" % tablespaceTable

        self.create["tp_threadpool_seq"] = \
          """CREATE SEQUENCE tp_threadpool_seq
               START WITH 1
               INCREMENT BY 1
               NOMAXVALUE"""

        self.indexes["tp_threadpool_pk"] = \
          """ALTER TABLE tp_threadpool ADD
               (CONSTRAINT tp_threadpool_pk PRIMARY KEY (id) %s)""" % tablespaceIndex

        self.constraints["tp_threadpool_fk"] = \
          """ALTER TABLE tp_threadpool ADD
               (CONSTRAINT tp_threadpool_state CHECK(state IN ('queued', 'process')))"""

        self.create["tp_threadpool_trg"] = \
          """CREATE TRIGGER tp_threadpool_trg
               BEFORE INSERT ON tp_threadpool
               FOR EACH ROW
                 DECLARE m_no INTEGER;
                 BEGIN
                   SELECT tp_threadpool_seq.nextval INTO :new.id FROM dual;
                 END;"""

        self.create["tp_threadpool_buffer_in"] = \
          """CREATE TABLE tp_threadpool_buffer_in (
               id             NUMBER(11)   NOT NULL ENABLE,
               event          VARCHAR(255) NOT NULL ENABLE,
               component      VARCHAR(255) NOT NULL ENABLE,
               payload        CLOB         NOT NULL ENABLE,
               thread_pool_id VARCHAR(255) NOT NULL ENABLE,
               state          VARCHAR2(20) DEFAULT 'queued' NOT NULL ENABLE
               ) %s""" % tablespaceTable

        self.create["tp_threadpool_buffer_in_seq"] = \
          """CREATE SEQUENCE tp_buffer_in_seq
               START WITH 1
               INCREMENT BY 1
               NOMAXVALUE"""

        self.create["tp_threadpool_buffer_in_trg"] = \
          """CREATE TRIGGER tp_buffer_in_trg
               BEFORE INSERT ON tp_threadpool_buffer_in
               FOR EACH ROW
                 BEGIN
                   SELECT tp_buffer_in_seq.nextval INTO :new.id FROM dual;
                 END;"""

        self.constraints["tp_threadpool_buffer_in_fk"] = \
          """ALTER TABLE tp_threadpool_buffer_in ADD
               (CONSTRAINT tp_threadpool_buffer_in_state CHECK(state IN ('queued', 'process')))"""

        self.indexes["tp_threadpool_buffer_in_pk"] = \
          """ALTER TABLE tp_threadpool_buffer_in ADD
               (CONSTRAINT tp_threadpool_buffer_in_pk PRIMARY KEY (id) %s)""" % tablespaceIndex

        self.create["tp_threadpool_buffer_out"] = \
          """CREATE TABLE tp_threadpool_buffer_out (
               id             NUMBER(11)   NOT NULL ENABLE,
               event          varchar(255) NOT NULL ENABLE,
               component      varchar(255) NOT NULL ENABLE,
               payload        clob         NOT NULL ENABLE,
               thread_pool_id varchar(255) NOT NULL ENABLE,
               state          varchar2(20) DEFAULT 'queued' NOT NULL ENABLE
               ) %s""" % tablespaceTable

        self.create["threadpool_buffer_out_seq"] = \
          """CREATE SEQUENCE tp_buffer_out_seq
               START WITH 1
               INCREMENT BY 1
               NOMAXVALUE"""

        self.create["tp_threadpool_buffer_out_trg"] = \
          """CREATE TRIGGER tp_buffer_out_trg
               BEFORE INSERT ON tp_threadpool_buffer_out
               FOR EACH ROW
                 BEGIN
                   SELECT tp_buffer_out_seq.nextval INTO :new.id FROM dual;
                 END;"""

        self.indexes["tp_threadpool_buffer_out_pk"] = \
          """ALTER TABLE tp_threadpool_buffer_out ADD
               (CONSTRAINT tp_threadpool_buffer_out_pk PRIMARY KEY (id) %s)""" % tablespaceIndex

        self.constraints["tp_threadpool_buffer_out_ck"] = \
          """ALTER TABLE tp_threadpool_buffer_out ADD
               (CONSTRAINT tp_threadpool_buffer_out_state CHECK(state IN ('queued', 'process')))"""



from __future__ import (division, print_function)

from builtins import str, bytes, object

from WMCore.Database.CMSCouch import CouchServer, Database
from WMCore.Lexicon import splitCouchServiceURL

class FWJRDBAPI(object):

    def __init__(self, couchURL, dbName=None):
        """
        setting up comon variables for inherited class.
        inherited class should call this in their init function
        """
        if isinstance(couchURL, Database):
            self.couchDB = couchURL
            self.couchURL = self.couchDB['host']
            self.dbName = self.couchDB.name
            self.couchServer = CouchServer(self.couchURL)
        else:
            if dbName == None:
                self.couchURL, self.dbName = splitCouchServiceURL(couchURL)
            else:
                self.couchURL = couchURL
                self.dbName = dbName
            self.couchServer = CouchServer(self.couchURL)
            self.couchDB = self.couchServer.connectDatabase(self.dbName, False)
        self.couchapp = "FWJRDump"
        self.defaultStale = {"stale": "update_after"}


    def setDefaultStaleOptions(self, options):
        if not options:
            options = {}
        if 'stale' not in options:
            options.update(self.defaultStale)
        return options

    def _setNoStale(self):
        """
        Use this only for the unittest
        """
        self.defaultStale = {}

    def _getCouchView(self, view, options, keys = []):

        options = self.setDefaultStaleOptions(options)

        if keys and isinstance(keys, (str, bytes)):
            keys = [keys]
        return self.couchDB.loadView(self.couchapp, view, options, keys)


    def _filterCouchInfo(self, couchInfo):
        # remove the couch specific information
        for key in ['_rev', '_attachments']:
            if  key in couchInfo:
                del couchInfo[key]
        return


    def _formatCouchData(self, data, key = "id", detail = True, filterCouch = True,
                         returnDict = False):
        result = {}
        for row in data['rows']:
            if 'error' in row:
                continue
            if "doc" in row:
                if filterCouch:
                    self._filterCouchInfo(row["doc"])
                result[row[key]] = row["doc"]
            else:
                result[row[key]] = row["value"]
        if detail or returnDict:
            return result
        else:
            return list(result)

    def getFWJRByArchiveStatus(self, status, limit=None, skip=None):
        """
        'status': list of the status or status string
        """
        options = {}
        options["include_docs"] = True

        if limit != None:
            options["limit"] = limit
        if skip != None:
            options["skip"] = skip
        keys = status
        return self._getCouchView("reportsByArchiveStatus", options, keys)

    def updateArchiveUploadedStatus(self, docID):

        return self.couchDB.updateDocument(docID, self.couchapp, "archiveStatus")

    def isAllFWJRArchived(self, workflow):
        keys = [[workflow, "ready"]]
        options = {"reduce": True, "group": True}
        result = self._getCouchView("byWorkflowAndArchiveStatus", options, keys)
        if len(result["rows"]) == 0:
            return True
        else:
            return False


    def outputByWorkflowName(self):

        options = {"group": True, "stale": "ok", "reduce": True}
        # site of data should be relatively small (~1M) for put in the memory
        # If not, find a way to stream
        return self._getCouchView("outputByWorkflowName", options)


    def getFWJRWithSkippedFiles(self):
        options = {"reduce": True, "group": True, "include_docs": False}
        return self._getCouchView("skippedFileInfoByTaskAndSite", options)




#!/usr/bin/env python
"""
_GetByLFN_

Oracle implementation of GetByLFN
"""




from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.GetByLFN import GetByLFN as MySQLGetByLFN

class GetByLFN(MySQLGetByLFN):
    """

    Oracle implementation of GetByLFN

    """
    pass



"""
_ListUserContent_

Implementation of ListUserContent for Oracle

"""

from WMCore.Database.DBFormatter import DBFormatter

class ListUserContent(DBFormatter):

    def execute(self, subscription = None, conn = None, transaction = False):

        sql = """SELECT object_name FROM user_objects"""

        result = self.dbi.processData(sql, {}, conn = conn,
                                      transaction = transaction)

        return self.formatDict(result)



#!/usr/bin/env python
"""
This script can be used to update the status of a request in ReqMgr2.
"""
from __future__ import print_function, division

from future import standard_library
standard_library.install_aliases()
import http.client
import json
import os
import sys


def setStatus(url, workflow, newstatus):
    headers = {"Content-type": "application/json",
               "Accept": "application/json"}

    print("Setting %s to %s" % (workflow, newstatus))
    if newstatus in ['closed-out', 'announced', 'aborted', 'rejected']:
        encodedParams = json.dumps({"RequestStatus": newstatus, "cascade": True})
    else:
        encodedParams = json.dumps({"RequestStatus": newstatus})

    conn = http.client.HTTPSConnection(url, cert_file=os.getenv('X509_USER_PROXY'), key_file=os.getenv('X509_USER_PROXY'))
    conn.request("PUT", "/reqmgr2/data/request/%s" % workflow, encodedParams, headers)
    resp = conn.getresponse()
    if resp.status != 200:
        print("Response status: %s\tResponse reason: %s" % (resp.status, resp.reason))
        if hasattr(resp.msg, "x-error-detail"):
            print("Error message: %s" % resp.msg["x-error-detail"])
            sys.exit(2)
    else:
        print("  OK!")
    conn.close()


def getStatus(url, workflow):
    headers = {"Content-type": "application/json",
               "Accept": "application/json"}

    conn = http.client.HTTPSConnection(url, cert_file=os.getenv('X509_USER_PROXY'), key_file=os.getenv('X509_USER_PROXY'))
    urn = "/reqmgr2/data/request/%s" % workflow
    conn.request("GET", urn, headers=headers)
    res = conn.getresponse()
    request = json.loads(res.read())["result"][0]
    return request[workflow]['RequestStatus']


def main():
    url = 'cmsweb.cern.ch'

    args = sys.argv[1:]
    if not len(args) == 2:
        print("usage: python setrequeststatus.py <text_file_with_the_workflow_names> <newStatus>")
        sys.exit(0)
    inputFile = args[0]
    newstatus = args[1]
    with open(inputFile, 'r') as fOjb:
        workflows = fOjb.readlines()

    for wflowName in workflows:
        wflowName = wflowName.rstrip('\n')
        print("%s" % wflowName)
        print("Set %s from %s to %s" % (wflowName, getStatus(url, wflowName), newstatus))
        setStatus(url, wflowName, newstatus)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
File       : MSUnmerged.py

Description:

This MicroService is meant to remove all no longer needed files from the unmerged
area of the CMS LFN Namespace. The cleanup process is supposed to happen on a per
RSE basis.
"""

# futures
from __future__ import division, print_function

from pprint import pformat
from time import time

import random
import re
import os
import errno
import stat
try:
    import gfal2
except ImportError:
    # in case we do not have gfal2 installed
    print("FAILED to import gfal2. Use it only in emulateGfal2=True mode!!!")
    gfal2 = None

from pymongo import IndexModel
from pymongo.errors  import NotPrimaryError

# WMCore modules
from WMCore.MicroService.DataStructs.DefaultStructs import UNMERGED_REPORT
from WMCore.MicroService.MSCore.MSCore import MSCore
from WMCore.MicroService.MSUnmerged.MSUnmergedRSE import MSUnmergedRSE
from WMCore.Services.RucioConMon.RucioConMon import RucioConMon
from WMCore.Services.WMStatsServer.WMStatsServer import WMStatsServer
from WMCore.Database.MongoDB import MongoDB
from WMCore.WMException import WMException
from Utils.Pipeline import Pipeline, Functor
from Utils.TwPrint import twFormat
from Utils.IteratorTools import grouper

# from memory_profiler import profile


class MSUnmergedException(WMException):
    """
    General Exception Class for MSUnmerged Module in WMCore MicroServices
    """
    def __init__(self, message=""):
        self.message = "MSUnmergedException: %s" % message
        super(MSUnmergedException, self).__init__(self.message)


class MSUnmergedPlineExit(MSUnmergedException):
    """
    An Exception Class for MSUnmerged to signal an expected pipeline exit condition
    This one should not produce any back trace dump.
    """
    def __init__(self, message=""):
        self.message = "MSUnmergedPlineExit: %s" % message
        super(MSUnmergedPlineExit, self).__init__(self.message)


def createGfal2Context(logLevel="normal", emulate=False):
    """
    Create a gfal2 context object
    :param logLevel: string with the gfal2 log level
    :param emulate: boolean to be used by unit tests
    :return: the gfal2 context object
    """
    if emulate:
        return None
    ctx = gfal2.creat_context()
    gfal2.set_verbose(gfal2.verbose_level.names[logLevel])
    return ctx


class MSUnmerged(MSCore):
    """
    MSUnmerged.py class provides the logic for cleaning the unmerged area of
    the CMS LFN Namespace.
    """

    # @profile
    def __init__(self, msConfig, logger=None):
        """
        Runs the basic setup and initialization for the MSUnmerged module
        :param msConfig: micro service configuration
        """
        super(MSUnmerged, self).__init__(msConfig, logger=logger)

        self.msConfig.setdefault("verbose", True)
        self.msConfig.setdefault("interval", 60)
        self.msConfig.setdefault("limitFilesPerRSE", 200)
        self.msConfig.setdefault("skipRSEs", [])
        self.msConfig.setdefault("rseExpr", "*")
        self.msConfig.setdefault("enableRealMode", False)
        self.msConfig.setdefault("enableT0WMStats", False)
        self.msConfig.setdefault("fullRSEToDB", False)
        self.msConfig.setdefault("gfalLogLevel", 'normal')
        self.msConfig.setdefault("dirFilterIncl", [])
        self.msConfig.setdefault("dirFilterExcl", [])
        self.msConfig.setdefault("emulateGfal2", False)
        self.msConfig.setdefault("filesToDeleteSliceSize", 100)

        self.msConfig.setdefault("mongoDBRetryCount", 3)
        self.msConfig.setdefault("mongoDBReplicaSet", None)
        self.msConfig.setdefault("mongoDBPort", None)
        self.msConfig.setdefault("mockMongoDB", False)

        msUnmergedIndex = IndexModel('name', unique=True)

        # NOTE: A full set of valid database connection parameters can be found at:
        #       https://pymongo.readthedocs.io/en/stable/api/pymongo/mongo_client.html
        msUnmergedDBConfig = {
            'database': self.msConfig['mongoDB'],
            'server': self.msConfig['mongoDBServer'],
            'port': self.msConfig['mongoDBPort'],
            'replicaSet': self.msConfig['mongoDBReplicaSet'],
            'username': self.msConfig['mongoDBUser'],
            'password': self.msConfig['mongoDBPassword'],
            'connect': True,
            'directConnection': False,
            'logger': self.logger,
            'create': True,
            'mockMongoDB': self.msConfig['mockMongoDB'],
            'collections': [('msUnmergedColl', msUnmergedIndex)]}

        mongoDB = MongoDB(**msUnmergedDBConfig)
        self.msUnmergedDB = getattr(mongoDB, self.msConfig['mongoDB'])
        self.msUnmergedColl = self.msUnmergedDB['msUnmergedColl']

        if self.msConfig['emulateGfal2'] is False and gfal2 is None:
            msg = "Failed to import gfal2 library while it's not "
            msg += "set to emulate it. Crashing the service!"
            raise ImportError(msg)

        # Instantiating the Rucio Consistency Monitor Client
        self.rucioConMon = RucioConMon(self.msConfig['rucioConMon'], logger=self.logger)

        self.wmstatsSvc = WMStatsServer(self.msConfig['wmstatsUrl'], logger=self.logger)
        self.wmstatsSvcT0 = WMStatsServer(self.msConfig['wmstatsUrlT0'], logger=self.logger)

        # Building all the Pipelines:
        pName = 'plineUnmerged'
        self.plineUnmerged = Pipeline(name=pName,
                                      funcLine=[Functor(self.getRSEFromMongoDB),
                                                Functor(self.updateRSETimestamps, start=True, end=False),
                                                Functor(self.consRecordAge),
                                                Functor(self.getUnmergedFiles),
                                                Functor(self.filterUnmergedFiles),
                                                Functor(self.getPfn),
                                                Functor(self.cleanRSE),
                                                Functor(self.updateServiceCounters),
                                                Functor(self.updateRSETimestamps, start=False, end=True),
                                                Functor(self.uploadRSEToMongoDB),
                                                Functor(self.purgeRseObj)])
        # Initialization service counters:
        self.plineCounters = {}
        self.rseTimestamps = {}

        # Initialization service common data structures:
        self.rseConsStats = {}
        self.protectedLFNs = set()

        # The basic /store/unmerged regular expression:
        self.regStoreUnmergedLfn = re.compile("^/store/unmerged/.*$")
        self.regStoreUnmergedPfn = re.compile("^.+/store/unmerged/.*$")

        # log msConfig
        self.logger.info("msConfig: %s", pformat(self.msConfig))

    # @profile
    def execute(self):
        """
        Executes the whole MSUnmerged logic
        :return: summary
        """
        # start threads in MSManager which should call this method
        summary = dict(UNMERGED_REPORT)

        # fetch the protectedLFNs list
        try:
            self.protectedLFNs = set(self.wmstatsSvc.getProtectedLFNs())
            # self.logger.debug("protectedLFNs: %s", pformat(self.protectedLFNs))
            if not self.protectedLFNs:
                msg = "Could not fetch the protectedLFNs list from Production WMStatServer. "
                msg += "Skipping the current run."
                self.logger.error(msg)
                return summary

            protectedLFNsT0 = set()
            if self.msConfig['enableT0WMStats']:
                msg = "WMStatsServer.getProtectedLFNs for T0 is not yet implemented."
                raise NotImplementedError(msg)
                # protectedLFNs = set(self.wmstatsSvcT0.getProtectedLFNs())
                # if not protectedLFNsT0:
                #     msg = "Could not fetch the protectedLFNs list from T0 WMStatServer. "
                #     msg += "Skipping the current run."
                #     self.logger.error(msg)
                #     return summary

            self.protectedLFNs = self.protectedLFNs | protectedLFNsT0

        except Exception as ex:
            msg = "Unknown exception while trying to fetch the protectedLFNs list from WMStatServer. Error: {}".format(str(ex))
            self.logger.exception(msg)
            self.updateReportDict(summary, "error", msg)
            return summary

        # refresh statistics on every poling cycle
        try:
            self.rseConsStats = self.rucioConMon.getRSEStats()
            # self.logger.debug("protectedLFNs: %s", pformat(self.protectedLFNs))

            if not self.rseConsStats:
                msg = "Could not fetch statistics from Rucio Consistency Monitor. "
                msg += "Skipping the current run."
                self.logger.error(msg)
                return summary
        except Exception as ex:
            msg = "Unknown exception while trying to fetch statistics from Rucio Consistency Monitor. Error: {}".format(str(ex))
            self.logger.exception(msg)
            self.updateReportDict(summary, "error", msg)
            return summary

        try:
            rseList = self.getRSEList()
            msg = "Retrieved list of %s RSEs: %s "
            msg += "Service set to process up to %s RSEs per instance."
            self.logger.info(msg, len(rseList), pformat(rseList), self.msConfig["limitRSEsPerInstance"])
            random.shuffle(rseList)
        except Exception as err:  # general error
            msg = "Unknown exception while trying to estimate the final list of RSEs to work on. Error: {}".format(str(err))
            self.logger.exception(msg)
            self.updateReportDict(summary, "error", msg)
            return summary

        try:
            totalNumRses, totalNumFiles, numRsesCleaned, numFilesDeleted = self._execute(rseList)
            msg = "\nTotal number of RSEs processed: %s."
            msg += "\nTotal number of files fetched from RucioConMon: %s."
            msg += "\nNumber of RSEs cleaned: %s."
            msg += "\nNumber of files deleted: %s."
            self.logger.info(msg,
                             totalNumRses,
                             totalNumFiles,
                             numRsesCleaned,
                             numFilesDeleted)
            self.updateReportDict(summary, "total_num_rses", totalNumRses)
            self.updateReportDict(summary, "total_num_files", totalNumFiles)
            self.updateReportDict(summary, "num_rses_cleaned", numRsesCleaned)
            self.updateReportDict(summary, "num_files_deleted", numFilesDeleted)
        except Exception as ex:
            msg = "Unknown exception while running MSUnmerged thread Error: {}".format(str(ex))
            self.logger.exception(msg)
            self.updateReportDict(summary, "error", msg)

        return summary

    # @profile
    def _execute(self, rseList):
        """
        Executes the MSUnmerged pipelines
        :param rseList: A list of RSEs to work on
        :return:        a tuple with:
                            total number of RSEs
                            total number of files found for deletion
                            number of RSEs cleaned
                            number of files deleted
        """

        pline = self.plineUnmerged
        self.resetServiceCounters()
        self.plineCounters[pline.name]['totalNumRses'] = len(rseList)

        for rseName in rseList:
            try:
                if rseName not in self.msConfig['skipRSEs']:
                    pline.run(MSUnmergedRSE(rseName))
                else:
                    msg = "%s: Run on RSE: %s is skipped due to a restriction set in msConfig. "
                    msg += "Will NOT retry until the RSE is removed from 'skipRSEs' list."
                    self.logger.info(msg, pline.name, rseName)
                    continue
            except MSUnmergedPlineExit as ex:
                msg = "%s: Run on RSE: %s was interrupted due to: %s "
                msg += "Will retry again in the next cycle."
                self.logger.warning(msg, pline.name, rseName, ex.message)
                continue
            except Exception as ex:
                msg = "%s: General error from pipeline. RSE: %s. Error: %s "
                msg += "Will retry again in the next cycle."
                self.logger.exception(msg, pline.name, rseName, str(ex))
                continue
        return self.plineCounters[pline.name]['totalNumRses'], \
            self.plineCounters[pline.name]['totalNumFiles'], \
            self.plineCounters[pline.name]['rsesCleaned'], \
            self.plineCounters[pline.name]['filesDeletedSuccess']

    # @profile
    def cleanRSE(self, rse):
        """
        The method to implement the actual deletion of files for an RSE.
        :param rse: MSUnmergedRSE object to be cleaned
        :return:    The MSUnmergedRSE object
        """

        # Create the gfal2 context object:
        try:
            ctx = createGfal2Context(self.msConfig['gfalLogLevel'], self.msConfig['emulateGfal2'])
        except Exception as ex:
            msg = "RSE: %s, Failed to create gfal2 Context object. " % rse['name']
            msg += "Skipping it in the current run."
            self.logger.exception(msg)
            raise MSUnmergedPlineExit(msg) from ex

        filesToDeleteCurrRSE = 0

        # Start cleaning one directory at a time:
        for dirLfn, fileLfnGen in rse['files']['toDelete'].items():
            if dirLfn in rse['dirs']['deletedSuccess']:
                self.logger.info("RSE: %s, dir: %s already successfully deleted.", rse['name'], dirLfn)
                continue

            if self.msConfig['limitFilesPerRSE'] < 0 or \
               filesToDeleteCurrRSE < self.msConfig['limitFilesPerRSE']:

                # Now we consume the rse['files']['toDelete'][dirLfn] generator
                # upon that no values will be left in it. In case we need it again
                # we will have to recreate the filter as we did in self.filterUnmergedFiles()
                pfnList = []
                if not rse['pfnPrefix']:
                    # Fall back to calling Rucio on a per directory basis for
                    # resolving the lfn to pfn mapping
                    dirPfn = self.rucio.getPFN(rse['name'], dirLfn, operation='delete')[dirLfn]
                    for fileLfn in fileLfnGen:
                        fileLfnSuffix = fileLfn.split(dirLfn)[1]
                        filePfn = dirPfn + fileLfnSuffix
                        pfnList.append(filePfn)
                else:
                    # Proceed with assembling the full filePfn out of the rse['pfnPrefix'] and the fileLfn
                    dirPfn = rse['pfnPrefix'] + dirLfn
                    for fileLfn in fileLfnGen:
                        filePfn = rse['pfnPrefix'] + fileLfn
                        pfnList.append(filePfn)

                filesToDeleteCurrRSE += len(pfnList)
                msg = "\nRSE: %s \nDELETING: %s."
                msg += "\nPFN list with: %s entries: \n%s"
                self.logger.debug(msg, rse['name'], dirLfn, len(pfnList), twFormat(pfnList, maxLength=4))

                if self.msConfig['enableRealMode']:
                    # execute the actual deletion in bulk - full list of files per directory

                    deletedSuccess = 0
                    for pfnSlice in list(grouper(pfnList, self.msConfig["filesToDeleteSliceSize"])):
                        try:
                            delResult = ctx.unlink(pfnSlice)
                            # Count all the successfully deleted files (if a deletion was
                            # successful a value of None is put in the delResult list):
                            self.logger.debug("RSE: %s, Dir: %s, delResult: %s",
                                              rse['name'], dirLfn, pformat(delResult))
                            for gfalErr in delResult:
                                if gfalErr is None:
                                    deletedSuccess += 1
                                else:
                                    errMessage = os.strerror(gfalErr.code)
                                    rse['counters']['gfalErrors'].setdefault(errMessage, 0)
                                    rse['counters']['gfalErrors'][errMessage] += 1
                        except Exception as ex:
                            msg = "Error while cleaning RSE: %s. "
                            msg += "Will retry in the next cycle. Err: %s"
                            self.logger.exception(msg, rse['name'], str(ex))

                    self.logger.info("RSE: %s, Dir: %s, filesDeletedSuccess: %s",
                                      rse['name'], dirLfn, deletedSuccess)
                    rse['counters']['filesDeletedSuccess'] += deletedSuccess

                    # Now clean the whole branch
                    self.logger.debug("Purging dirEntry: %s:\n", dirPfn)
                    purgeSuccess = self._purgeTree(ctx, dirPfn)
                    if purgeSuccess:
                        rse['dirs']['deletedSuccess'].add(dirLfn)
                        rse['counters']['dirsDeletedSuccess'] = len(rse['dirs']['deletedSuccess'])
                        # if dirLfn in rse['dirs']['toDelete']:
                        #     rse['dirs']['toDelete'].remove(dirLfn)
                        if dirLfn in rse['dirs']['deletedFail']:
                            rse['dirs']['deletedFail'].remove(dirLfn)
                        msg = "RSE: %s  Success deleting directory: %s"
                        self.logger.info(msg, rse['name'], dirPfn)
                    else:
                        rse['dirs']['deletedFail'].add(dirLfn)
                        rse['counters']['dirsDeletedFail'] = len(rse['dirs']['deletedFail'])
                        msg = "RSE: %s Failed to purge directory: %s"
                        self.logger.error(msg, rse['name'], dirPfn)
            else:
                msg = "RSE: %s reached limit of files per RSE to be deleted. Skipping directory: %s. It will be retried on the next cycle."
                self.logger.warning(msg, rse['name'], dirLfn)
        rse['isClean'] = self._checkClean(rse)

        return rse

    def _purgeTree(self, ctx, baseDirPfn):
        """
        A method to be used for purging the tree bellow a specific branch.
        It deletes every empty directory bellow that branch + the origin at the end.
        :param ctx:  The gfal2 context object
        :return:     Bool: True if it managed to purge everything, False otherwise
        """
        # NOTE: It deletes only directories and does not try to unlink any file.

        # First test if baseDirPfn is actually a directory entry:
        try:
            entryStat = ctx.stat(baseDirPfn)
            if not stat.S_ISDIR(entryStat.st_mode):
                self.logger.error("The base pfn: %s is not a directory entry.", baseDirPfn)
                return False
        except gfal2.GError as gfalExc:
            if gfalExc.code == errno.ENOENT:
                self.logger.warning("MISSING baseDir: %s", baseDirPfn)
                return True
            else:
                self.logger.error("FAILED to open baseDir: %s: gfalException: %s", baseDirPfn, str(gfalExc))
                return False

        if baseDirPfn[-1] != '/':
            baseDirPfn += '/'

        # Second recursively iterate down the tree:
        successList = []
        for dirEntry in ctx.listdir(baseDirPfn):
            if dirEntry in ['.', '..']:
                continue
            dirEntryPfn = baseDirPfn + dirEntry
            try:
                entryStat = ctx.stat(dirEntryPfn)
            except gfal2.GError as gfalExc:
                if gfalExc.code == errno.ENOENT:
                    self.logger.warning("MISSING dirEntry: %s", dirEntryPfn)
                    successList.append(True)
                else:
                    self.logger.error("FAILED to open dirEntry: %s: gfalException: %s", dirEntryPfn, str(gfalExc))
                    successList.append(False)

            if stat.S_ISDIR(entryStat.st_mode):
                successList.append(self._purgeTree(ctx, dirEntryPfn))

        # Finally remove the baseDir:
        try:
            self.logger.debug("RM baseDir: %s", baseDirPfn)
            success = ctx.rmdir(baseDirPfn)
            # for gfal2 rmdir() exit status of 0 is success
            if success == 0:
                successList.append(True)
            else:
                successList.append(False)
        except gfal2.GError as gfalExc:
            if gfalExc.code == errno.ENOENT:
                self.logger.warning("MISSING baseDir: %s", baseDirPfn)
                successList.append(True)
            else:
                self.logger.error("FAILED to remove baseDir: %s: gfalException: %s", baseDirPfn, str(gfalExc))
                successList.append(False)
        return all(successList)

    def _checkClean(self, rse):
        """
        A simple function to check if every file in an RSE's unmerged area have
        been deleted
        :param rse: The RSE to be checked
        :return:    Bool: True if all files found have been deleted, False otherwise
        """
        return rse['counters']['dirsToDelete'] == rse['counters']['dirsDeletedSuccess']

    def consRecordAge(self, rse):
        """
        A method to check the duration of the consistency record for the RSE
        :param rse: The RSE to be checked
        :return:    rse or raises MSUnmergedPlineExit
        """
        rseName = rse['name']

        if rseName not in self.rseConsStats:
            msg = "RSE: %s Missing in stats records at Rucio Consistency Monitor. " % rseName
            msg += "Skipping it in the current run."
            self.logger.warning(msg)
            rse['rucioConMonStatus'] = "Missing"
            self.updateRSETimestamps(rse, start=False, end=True)
            self.uploadRSEToMongoDB(rse)
            raise MSUnmergedPlineExit(msg)

        isConsDone = self.rseConsStats[rseName]['status'] == 'done'
        if not isConsDone:
            msg = "RSE: %s has a non-final Rucio ConMon status: %s. " % (rseName, self.rseConsStats[rseName]['status'])
            msg += "Skipping it in the current run."
            self.logger.warning(msg)
            rse['rucioConMonStatus'] = self.rseConsStats[rseName]['status']
            self.updateRSETimestamps(rse, start=False, end=True)
            self.uploadRSEToMongoDB(rse)
            raise MSUnmergedPlineExit(msg)

        isConsNewer = self.rseConsStats[rseName]['end_time'] > self.rseTimestamps[rseName]['prevStartTime']
        if not isConsNewer:
            msg = "RSE: %s With old consistency record in Rucio Consistency Monitor. " % rseName
            if 'isClean' in rse and rse['isClean']:
                msg += "And the RSE has been cleaned during the last Rucio Consistency Monitor polling cycle."
                msg += "Skipping it in the current run."
                self.logger.info(msg)
                rse['rucioConMonStatus'] = self.rseConsStats[rseName]['status']
                self.updateRSETimestamps(rse, start=False, end=True)
                self.uploadRSEToMongoDB(rse)
                raise MSUnmergedPlineExit(msg)
            else:
                msg += "But the RSE has NOT been fully cleaned during the last Rucio Consistency Monitor polling cycle."
                msg += "Retrying cleanup in the current run."
                self.logger.info(msg)
                rse['rucioConMonStatus'] = self.rseConsStats[rseName]['status']

        if isConsNewer and isConsDone:
            # NOTE: If we've got to this point then we have a brand new record for
            #       the RSE in RucioConMOn and we are then about to start a new RSE cleanup
            #       so we will need to wipe out all but the timestamps from both
            #       the current object and the MongoDB record for the object.
            msg = "RSE: %s With new consistency record in Rucio Consistency Monitor. " % rseName
            msg += "Resetting RSE and starting a fresh cleanup process in the current run."
            self.logger.info(msg)
            try:
                rse.resetRSE(self.msUnmergedColl, keepTimestamps=True, retryCount=self.msConfig['mongoDBRetryCount'])
            except NotPrimaryError:
                msg = "Could not reset RSE to MongoDB for the maximum of %s mongoDBRetryCounts configured." % self.msConfig['mongoDBRetryCount']
                msg += "Giving up now. The whole cleanup process will be retried for this RSE on the next run."
                msg += "Duplicate deletion retries may cause error messages from false positives and wrong counters during next polling cycle."
                raise MSUnmergedPlineExit(msg) from None

        # Before returning the RSE update Consistency StatTime:
        rse['timestamps']['rseConsStatTime'] = self.rseConsStats[rseName]['end_time']
        return rse

    # @profile
    def getUnmergedFiles(self, rse):
        """
        Fetches all the records of unmerged files per RSE from Rucio Consistency Monitor
        and cuts everything to a certain level in the path and puts the list in the rse obj.

        Path example:
        /store/unmerged/Run2016B/JetHT/MINIAOD/ver2_HIPM_UL2016_MiniAODv2-v2/140000/388E3DEF-9F15-D04C-B582-7DD036D9DD33.root

        Where:
        /store/unmerged/                       - root unmerged area
        /Run2016B                              - acquisition era
        /JetHT                                 - primary dataset
        /MINIAOD                               - data tier
        /ver2_HIPM_UL2016_MiniAODv2-v2         - processing string + processing version
        /140000/388E3DEF-...-7DD036D9DD33.root - to be cut off

        :param rse: The RSE to work on
        :return:    rse
        """
        rse['files']['allUnmerged'] = self.rucioConMon.getRSEUnmerged(rse['name'])
        for filePath in rse['files']['allUnmerged']:
            # Check if what we start with is under /store/unmerged/*
            if self.regStoreUnmergedLfn.match(filePath):
                # Cut the path to the deepest level known to WMStats protected LFNs
                dirPath = self._cutPath(filePath)
                # Check if what is left is still under /store/unmerged/*
                if self.regStoreUnmergedLfn.match(dirPath):
                    # Add it to the set of allUnmerged
                    rse['dirs']['allUnmerged'].add(dirPath)

        rse['counters']['totalNumFiles'] = len(rse['files']['allUnmerged'])
        rse['counters']['totalNumDirs'] = len(rse['dirs']['allUnmerged'])
        return rse

    def _cutPath(self, filePath):
        """
        Cuts a file path to the deepest level known to WMStats protected LFNs
        :param filePath:   The full (absolute) file path together with the file name
        :return finalPath: The final path cut the to correct level
        """
        # pylint: disable=E1120
        # This is a known issue when when passing an unpacked list to a method expecting
        # at least one variable. In this case the signature of the method breaking the
        # rule is:
        # os.path.join(*newPath) != os.path.join(a, *p)

        # Split the initial filePath into chunks and fill it into a dictionary
        # containing only directory names and the root of the path e.g.
        # ['/', 'store', 'unmerged', 'RunIISummer20UL17SIM', ...]
        newPath = []
        root = filePath
        while True:
            root, tail = os.path.split(root)
            if tail:
                newPath.append(tail)
            else:
                newPath.append(root)
                break
        newPath.reverse()
        # Cut/slice the path to the level/element required.
        newPath = newPath[:7]
        # Build the path out of all that is found up to the deepest level in the LFN tree
        finalPath = os.path.join(*newPath)
        return finalPath

    # @profile
    def filterUnmergedFiles(self, rse):
        """
        This method is applying set compliment operation to the set of unmerged
        files per RSE in order to exclude the protected LFNs.
        :param rse: The RSE to work on
        :return:    rse
        """
        rse['dirs']['toDelete'] = rse['dirs']['allUnmerged'] - self.protectedLFNs
        rse['dirs']['protected'] = rse['dirs']['allUnmerged'] & self.protectedLFNs

        # The following check may seem redundant, but better stay safe than sorry
        if not (rse['dirs']['toDelete'] | rse['dirs']['protected']) == rse['dirs']['allUnmerged']:
            rse['counters']['dirsToDelete'] = -1
            msg = "Incorrect set check while trying to estimate the final set for deletion."
            raise MSUnmergedPlineExit(msg)

        # Get rid of 'allUnmerged' directories
        rse['dirs']['allUnmerged'].clear()

        # NOTE: Here we may want to filter out all protected files from allUnmerged and leave just those
        #       eligible for deletion. This will minimize the iteration time of the filters
        #       from toDelete later on.
        # while rse['files']['allUnmerged'

        # Now create the filters for rse['files']['toDelete'] - those should be pure generators
        # A simple generator:
        def genFunc(pattern, iterable):
            for i in iterable:
                if i.startswith(pattern):
                    yield i

        # NOTE: If the 'dirFilterIncl' is non empty then the cleaning process will
        #       be enclosed only in this part of the tree and will ignore anything
        #       from /store/unmerged/ which does not belong to the included filter
        # NOTE: 'dirFilterExcl' is always applied.

        # Merge the additional filters into a final set to be applied:
        dirFilterIncl = set(self.msConfig['dirFilterIncl'])
        dirFilterExcl = set(self.msConfig['dirFilterExcl'])

        # Update directory/files with no service filters
        if not dirFilterIncl and not dirFilterExcl:
            for dirName in rse['dirs']['toDelete']:
                rse['files']['toDelete'][dirName] = genFunc(dirName, rse['files']['allUnmerged'])
            rse['counters']['dirsToDelete'] = len(rse['files']['toDelete'])
            self.logger.info("RSE: %s: %s", rse['name'], twFormat(rse, maxLength=8))
            return rse

        # If we are here, then there are service filters...
        for dirName in rse['dirs']['toDelete']:
            # apply exclusion filter
            dirFilterExclMatch = []
            for pathExcl in dirFilterExcl:
                dirFilterExclMatch.append(dirName.startswith(pathExcl))
            if any(dirFilterExclMatch):
                # then it matched one of the exclusion paths
                continue
            if not dirFilterIncl:
                # there is no inclusion filter, simply add this directory/files
                rse['files']['toDelete'][dirName] = genFunc(dirName, rse['files']['allUnmerged'])
                continue

            # apply inclusion filter
            for pathIncl in dirFilterIncl:
                if dirName.startswith(pathIncl):
                    rse['files']['toDelete'][dirName] = genFunc(dirName, rse['files']['allUnmerged'])
                    break

        # Now apply the filters back to the set in rse['dirs']['toDelete']
        rse['dirs']['toDelete'] = set(rse['files']['toDelete'].keys())

        # Update the counters:
        rse['counters']['dirsToDelete'] = len(rse['files']['toDelete'])
        self.logger.info("RSE: %s: %s", rse['name'], twFormat(rse, maxLength=8))
        return rse

    def getPfn(self, rse):
        """
        A method for fetching the common Pfn (method + hostname + global path)
        for the RSE. It uses Rucio client method lfns2pfns for one of the LFNs
        already recorded in the RSE in order to resolve the lfn to pfn mapping
        and then tries to parse the resultant pfn and cut off the lfn part.
        :param rse: The RSE to be checked
        :return:    rse
        """
        # NOTE:  pfnPrefix here is considered the full part of the pfn up to the
        #        beginning of the lfn part rather than just the protocol prefix
        if rse['files']['allUnmerged']:
            lfn = next(iter(rse['files']['allUnmerged']))
            pfnDict = self.rucio.getPFN(rse['name'], lfn, operation='delete')
            pfnFull = pfnDict[lfn]
            if self.regStoreUnmergedPfn.match(pfnFull):
                pfnPrefix = pfnFull.split('/store/unmerged/')[0]
                rse['pfnPrefix'] = pfnPrefix
            else:
                msg = "Could not establish the correct pfn Prefix for RSE: %s. " % rse['name']
                msg += "Will fall back to calling Rucio on a directory basis for lfn to pfn resolution."
                self.logger.warning(msg)
        return rse

    # @profile
    def purgeRseObj(self, rse, dumpRSEtoLog=False):
        """
        Cleaning all the records in an RSE object. The final method to be used
        before an RSE exits a pipeline.
        :param rse: The RSE to be checked
        :param dumpRSEToLog: Dump the whole RSEobject into the service log.
        :return:    rse
        """
        msg = "\n----------------------------------------------------------"
        msg += "\nMSUnmergedRSE: \n%s"
        msg += "\n----------------------------------------------------------"
        if dumpRSEtoLog:
            self.logger.debug(msg, pformat(rse))
        else:
            self.logger.debug(msg, twFormat(rse, maxLength=8))
        rse.clear()
        return rse

    def updateRSETimestamps(self, rse, start=True, end=True):
        """
        Update/Upload all timestamps for the rse object into the MSUnmerged
        service counters
        :param rse:   The RSE to work on
        :return:      rse
        """
        rseName = rse['name']
        currTime = time()

        # Initialize the timestamps in both MSUnmerged and MSUnmergedRSE objects
        if rseName not in self.rseTimestamps:
            # NOTE: Reading the timestamps from the rse for the first time.
            #       This will reset them if no previous record for the RSE in MongoDB
            #       or will set them to the values fetched from MongoDB, provided
            #       that the RSE object has been updated from the database.
            self.rseTimestamps[rseName] = rse['timestamps']

        # # Read last RucioConMon stat time for this RSE:
        # self.rseTimestamps[rseName]['rseConsStatsTime'] = self.rseConsStats[rseName]['startTime']

        # Update the timestamps:
        if start:
            self.rseTimestamps[rseName]['prevStartTime'] = self.rseTimestamps[rseName]['startTime']
            self.rseTimestamps[rseName]['startTime'] = currTime
        if end:
            self.rseTimestamps[rseName]['prevEndTime'] = self.rseTimestamps[rseName]['endTime']
            self.rseTimestamps[rseName]['endTime'] = currTime
        rse['timestamps'] = self.rseTimestamps[rseName]
        return rse

    def updateServiceCounters(self, rse):
        """
        Update/Upload all counters from the rse object into the MSUnmerged
        service counters
        :param rse:   The RSE to work on
        :param pName: The pipeline name whose counters to be updated
        :return:      rse
        """
        pName = self.plineUnmerged.name
        self.plineCounters[pName]['totalNumFiles'] += rse['counters']['totalNumFiles']
        self.plineCounters[pName]['totalNumDirs'] += rse['counters']['totalNumDirs']
        self.plineCounters[pName]['filesDeletedSuccess'] += rse['counters']['filesDeletedSuccess']
        self.plineCounters[pName]['filesDeletedFail'] += rse['counters']['filesDeletedFail']
        self.plineCounters[pName]['dirsDeletedSuccess'] += rse['counters']['dirsDeletedSuccess']
        self.plineCounters[pName]['dirsDeletedFail'] += rse['counters']['dirsDeletedFail']
        self.plineCounters[pName]['rsesProcessed'] += 1
        if rse['isClean']:
            self.plineCounters[pName]['rsesCleaned'] += 1

        return rse

    def resetServiceCounters(self):
        """
        A simple function for zeroing the service counters.
        :param plineName: The Pline Name whose counters to be zeroed
        """
        # Resetting pline counters
        plineName = self.plineUnmerged.name
        self.plineCounters.setdefault(plineName, {})
        self.plineCounters[plineName]['totalNumFiles'] = 0
        self.plineCounters[plineName]['totalNumDirs'] = 0
        self.plineCounters[plineName]['filesDeletedSuccess'] = 0
        self.plineCounters[plineName]['filesDeletedFail'] = 0
        self.plineCounters[plineName]['dirsDeletedSuccess'] = 0
        self.plineCounters[plineName]['dirsDeletedFail'] = 0
        self.plineCounters[plineName]['totalNumRses'] = 0
        self.plineCounters[plineName]['rsesProcessed'] = 0
        self.plineCounters[plineName]['rsesCleaned'] = 0
        return

    def getRSEFromMongoDB(self, rse):
        """
        Updates the record for an RSE from MongoDB
        :param rse: The RSE object to work on
        :return:    rse
        """
        self.logger.info("RSE: %s Reading rse data from MongoDB.", rse['name'])
        rse.readRSEFromMongoDB(self.msUnmergedColl)
        return rse

    def uploadRSEToMongoDB(self, rse, fullRSEToDB=False, overwrite=True):
        """
        Updates the record for an RSE at MongoDB
        :param rse: The RSE object to work on
        :return:    rse
        """
        try:
            self.logger.info("RSE: %s Writing rse data to MongoDB.", rse['name'])
            rse.writeRSEToMongoDB(self.msUnmergedColl, fullRSEToDB=fullRSEToDB, overwrite=overwrite, retryCount=self.msConfig['mongoDBRetryCount'])
        except NotPrimaryError:
            msg = "Could not write RSE to MongoDB for the maximum of %s mongoDBRetryCounts configured." % self.msConfig['mongoDBRetryCount']
            msg += "Giving up now. The whole cleanup process will be retried for this RSE on the next run."
            msg += "Duplicate deletion retries may cause error messages from false positives and wrong counters during next polling cycle."
            raise MSUnmergedPlineExit(msg) from None
        return rse

    def getStatsFromMongoDB(self, detail=False, **kwargs):
        """
        Auxiliary method to serve the APIs of the REST calls for the service.
        Implements various queries to MongoDB based on the parameters passed.
        :param detail:           Bool marking if additional details must be queried from the database
                                 (e.g. putting the 'dirs' field into the mongo projection)
        :param rse=rseName:      String representing the RSE name to query the database for
        :return:                 Dictionary with the type of query + the data returned from the database
        """
        data = {}
        if kwargs.get('rse'):
            data["query"] = 'rse=%s&detail=%s' % (kwargs['rse'], detail)
            allDocs = (kwargs['rse'].lower() == "all_docs") if isinstance(kwargs['rse'], str) else False

            if allDocs:
                mongoProjection = {
                    "_id": False,
                    "name": True,
                    "isClean": True,
                    "rucioConMonStatus": True,
                    "counters": {
                        "gfalErrors": True,
                        "dirsToDelete": True,
                        "dirsDeletedSuccess": True,
                        "dirsDeletedFail": True}}
                mongoFilter = {}
                data["rseData"] = list(self.msUnmergedColl.find(mongoFilter, projection=mongoProjection))
            else:
                mongoProjection = {
                    "_id": False,
                    "name": True,
                    "isClean": True,
                    "pfnPrefix": True,
                    "rucioConMonStatus": True,
                    "timestamps": True,
                    "counters": True}
                if detail:
                    mongoProjection["dirs"]  = True

                rseList = kwargs['rse'] if isinstance(kwargs['rse'], list) else [kwargs['rse']]
                data["rseData"] = []
                for rseName in rseList:
                    mongoFilter = {'name': rseName}
                    data["rseData"].append(self.msUnmergedColl.find_one(mongoFilter, projection=mongoProjection))
        return data

    # @profile
    def getRSEList(self):
        """
        Queries Rucio for the proper RSE list to iterate through.
        :return: List of RSE names.
        """
        try:
            rseList = self.rucio.evaluateRSEExpression(self.msConfig['rseExpr'])
        except Exception as ex:
            msg = "Unknown exception while trying to fetch the initial list of RSEs to work on. Err: %s"
            self.logger.exception(msg, str(ex))
            rseList = []
        return rseList



"""
_JobTypeCountByState_

Monitoring DAO classes for Jobs in WMBS
"""
from __future__ import print_function, division

from WMCore.Database.DBFormatter import DBFormatter


class JobTypeCountByState(DBFormatter):

    sqlType = "SELECT name FROM wmbs_sub_types"

    sql = """
          SELECT wmbs_sub_types.name AS job_type, COUNT(*) AS count FROM wmbs_job
            INNER JOIN wmbs_jobgroup ON wmbs_job.jobgroup = wmbs_jobgroup.id
            INNER JOIN wmbs_subscription ON wmbs_jobgroup.subscription = wmbs_subscription.id
            INNER JOIN wmbs_sub_types ON wmbs_subscription.subtype = wmbs_sub_types.id
            WHERE wmbs_job.state = (SELECT id FROM wmbs_job_state WHERE name = :state)
            GROUP BY wmbs_subscription.subtype, wmbs_sub_types.name
          """

    def __init__(self, logger, dbinterface):
        super(JobTypeCountByState, self).__init__(logger, dbinterface)
        self.jobTypes = []

    def formatDict(self, results):
        """
        _formatDict_

        Format the results in a dict keyed by the state name.
        Also add job/sub types not returned by the query, for consistency.
        """
        formattedResults = DBFormatter.formatDict(self, results)

        dictResult = {}
        for formattedResult in formattedResults:
            dictResult[formattedResult['job_type']] = int(formattedResult['count'])

        for jType in self.jobTypes:
            dictResult.setdefault(jType, 0)

        return dictResult

    def execute(self, state, conn=None, transaction=False):
        # first fetch all job types available in the agent and keep it in the object
        if not self.jobTypes:
            result = self.dbi.processData(self.sqlType, conn=conn, transaction=transaction)
            self.jobTypes = self.formatList(result)

        binds = {'state': state}
        result = self.dbi.processData(self.sql, binds, conn=conn, transaction=transaction)

        return self.formatDict(result)



"""
Gets the cache data from server cache. This shouldn't update the server cache.
Just wait for the server cache to be updated
"""
from __future__ import (division, print_function)
from WMCore.REST.Server import RESTEntity, restcall, rows
from WMCore.REST.Tools import tools
from WMCore.REST.Error import DataCacheEmpty
from WMCore.WMStats.DataStructs.DataCache import DataCache
from WMCore.REST.Format import JSONFormat, PrettyJSONFormat
from WMCore.ReqMgr.DataStructs.RequestStatus import ACTIVE_STATUS_FILTER


class ActiveRequestJobInfo(RESTEntity):
    """
    get all the active requests with job information attatched
    """

    def __init__(self, app, api, config, mount):
        # main CouchDB database where requests/workloads are stored
        RESTEntity.__init__(self, app, api, config, mount)

    def validate(self, apiobj, method, api, param, safe):
        return

    @restcall(formats=[('text/plain', PrettyJSONFormat()), ('application/json', JSONFormat())])
    @tools.expires(secs=-1)
    def get(self):
        # This assumes DataCahe is periodically updated.
        # If data is not updated, need to check, dataCacheUpdate log
        return rows([DataCache.getlatestJobData()])


class FilteredActiveRequestJobInfo(RESTEntity):
    """
    get all the active requests with job information attatched
    """

    def __init__(self, app, api, config, mount):
        # main CouchDB database where requests/workloads are stored
        RESTEntity.__init__(self, app, api, config, mount)

    def validate(self, apiobj, method, api, param, safe):

        for prop in param.kwargs:
            safe.kwargs[prop] = param.kwargs[prop]

        for prop in safe.kwargs:
            del param.kwargs[prop]

        return

    @restcall(formats=[('text/plain', PrettyJSONFormat()), ('application/json', JSONFormat())])
    @tools.expires(secs=-1)
    def get(self, mask=None, **input_condition):
        # This assumes DataCahe is periodically updated.
        # If data is not updated, need to check, dataCacheUpdate log
        return rows(DataCache.filterDataByRequest(input_condition, mask))


class ProtectedLFNList(RESTEntity):
    """
    API which provides a list of ALL possible unmerged LFN bases (including
    transient datasets/LFNs).
    """

    def __init__(self, app, api, config, mount):
        # main CouchDB database where requests/workloads are stored
        RESTEntity.__init__(self, app, api, config, mount)

    def validate(self, apiobj, method, api, param, safe):
        return

    @restcall(formats=[('text/plain', PrettyJSONFormat()), ('application/json', JSONFormat())])
    @tools.expires(secs=-1)
    def get(self):
        # This assumes DataCahe is periodically updated.
        # If data is not updated, need to check, dataCacheUpdate log
        if DataCache.isEmpty():
            raise DataCacheEmpty()
        else:
            return rows(DataCache.filterData(ACTIVE_STATUS_FILTER, ["OutputModulesLFNBases"]))


class ProtectedLFNListOnlyFinalOutput(RESTEntity):
    """
    Same as ProtectedLFNList API, however this one only provides LFNs that are not
    transient, so only final output LFNs.
    """

    def __init__(self, app, api, config, mount):
        # main CouchDB database where requests/workloads are stored
        RESTEntity.__init__(self, app, api, config, mount)

    def validate(self, apiobj, method, api, param, safe):
        return

    @restcall(formats=[('text/plain', PrettyJSONFormat()), ('application/json', JSONFormat())])
    @tools.expires(secs=-1)
    def get(self):
        # This assumes DataCahe is periodically updated.
        # If data is not updated, need to check, dataCacheUpdate log
        return rows(DataCache.getProtectedLFNs())


class GlobalLockList(RESTEntity):
    def __init__(self, app, api, config, mount):
        # main CouchDB database where requests/workloads are stored
        RESTEntity.__init__(self, app, api, config, mount)

    def validate(self, apiobj, method, api, param, safe):
        return

    @restcall(formats=[('text/plain', PrettyJSONFormat()), ('application/json', JSONFormat())])
    @tools.expires(secs=-1)
    def get(self):
        # This assumes DataCahe is periodically updated.
        # If data is not updated, need to check, dataCacheUpdate log
        if DataCache.isEmpty():
            raise DataCacheEmpty()
        else:
            return rows(DataCache.filterData(ACTIVE_STATUS_FILTER,
                                             ["InputDataset", "OutputDatasets", "MCPileup", "DataPileup"]))



#!/usr/bin/env python
"""
_EventBased_t_

Event based splitting test.
"""
from __future__ import print_function

from builtins import range
import unittest

from WMCore.DataStructs.File import File
from WMCore.DataStructs.Fileset import Fileset
from WMCore.DataStructs.Run import Run
from WMCore.DataStructs.Subscription import Subscription
from WMCore.DataStructs.Workflow import Workflow
from WMCore.JobSplitting.SplitterFactory import SplitterFactory
from WMCore.Services.UUIDLib import makeUUID


class EventBasedTest(unittest.TestCase):
    """
    _EventBasedTest_

    Test event based job splitting.
    """

    def setUp(self):
        """
        _setUp_

        Create two subscriptions: One that contains a single file and one that
        contains multiple files.
        """
        self.multipleFileFileset = Fileset(name="TestFileset1")
        for i in range(10):
            newFile = File(makeUUID(), size=1000, events=100)
            newFile.setLocation('se01')
            self.multipleFileFileset.addFile(newFile)

        self.singleFileFileset = Fileset(name="TestFileset2")
        newFile = File("/some/file/name", size=1000, events=100)
        newFile.setLocation('se02')
        self.singleFileFileset.addFile(newFile)

        self.emptyFileFileset = Fileset(name="TestFileset3")
        newFile = File("/some/file/name", size=1000, events=0)
        newFile.setLocation('se03')
        self.emptyFileFileset.addFile(newFile)

        testWorkflow = Workflow()
        self.multipleFileSubscription = Subscription(fileset=self.multipleFileFileset,
                                                     workflow=testWorkflow,
                                                     split_algo="EventBased",
                                                     type="Processing")
        self.singleFileSubscription = Subscription(fileset=self.singleFileFileset,
                                                   workflow=testWorkflow,
                                                   split_algo="EventBased",
                                                   type="Processing")
        self.emptyFileSubscription = Subscription(fileset=self.emptyFileFileset,
                                                  workflow=testWorkflow,
                                                  split_algo="EventBased",
                                                  type="Processing")

        self.eventsPerJob = 100
        self.performanceParams = {'timePerEvent': None,
                                  'memoryRequirement': 2300,
                                  'sizePerEvent': 400}

        return

    def tearDown(self):
        """
        _tearDown_

        Nothing to do...
        """
        pass

    def generateFakeMCFile(self, numEvents=100, firstEvent=1, lastEvent=100,
                           firstLumi=1, lastLumi=10, existingSub=None):
        # MC comes with only one MCFakeFile
        newFile = File("MCFakeFileTest", size=1000, events=numEvents)
        newFile.setLocation('se01')
        if firstLumi == lastLumi:
            newFile.addRun(Run(1, *list(range(firstLumi, lastLumi + 1))))
        else:
            newFile.addRun(Run(1, *list(range(firstLumi, lastLumi))))
        newFile["first_event"] = firstEvent
        newFile["last_event"] = lastEvent

        if existingSub is None:
            singleMCFileset = Fileset(name="MCTestFileset")
            singleMCFileset.addFile(newFile)
            testWorkflow = Workflow()
            existingSub = Subscription(fileset=singleMCFileset,
                                       workflow=testWorkflow,
                                       split_algo="EventBased",
                                       type="Production")
        else:
            existingSub['fileset'].addFile(newFile)

        return existingSub

    def testNoEvents(self):
        """
        _testNoEvents_

        Test event based job splitting where there are no events in the
        input file, make sure the mask events are None
        """
        splitter = SplitterFactory()
        jobFactory = splitter(self.emptyFileSubscription)
        jobGroups = jobFactory(events_per_job=100,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1)
        self.assertEqual(len(jobGroups[0].jobs), 1)

        for job in jobGroups[0].jobs:
            self.assertEqual(job.getFiles(type="lfn"), ["/some/file/name"])
            self.assertEqual(job["mask"].getMaxEvents(), None)
            self.assertEqual(job["mask"]["FirstEvent"], 0)
            self.assertEqual(job["mask"]["LastEvent"], None)

    def testExactEvents(self):
        """
        _testExactEvents_

        Test event based job splitting when the number of events per job is
        exactly the same as the number of events in the input file.
        """
        splitter = SplitterFactory()
        jobFactory = splitter(self.singleFileSubscription)
        jobGroups = jobFactory(events_per_job=self.eventsPerJob,
                               performance=self.performanceParams)

        assert len(jobGroups) == 1, \
            "ERROR: JobFactory didn't return one JobGroup."

        assert len(jobGroups[0].jobs) == 1, \
            "ERROR: JobFactory didn't create a single job."

        job = jobGroups[0].jobs.pop()

        assert job.getFiles(type="lfn") == ["/some/file/name"], \
            "ERROR: Job contains unknown files."

        self.assertEqual(job["mask"].getMaxEvents(), self.eventsPerJob, "ERROR: Job's max events is incorrect.")

        assert job["mask"]["FirstEvent"] == 0, \
            "ERROR: Job's first event is incorrect."

        return

    def testMoreEvents(self):
        """
        _testMoreEvents_

        Test event based job splitting when the number of events per job is
        greater than the number of events in the input file.
        """
        splitter = SplitterFactory()
        jobFactory = splitter(self.singleFileSubscription)

        jobGroups = jobFactory(events_per_job=1000,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1)

        self.assertEqual(len(jobGroups[0].jobs), 1)

        for job in jobGroups[0].jobs:
            self.assertEqual(job.getFiles(type="lfn"), ["/some/file/name"])
            self.assertEqual(job["mask"].getMaxEvents(), self.eventsPerJob)
            self.assertEqual(job["mask"]["FirstEvent"], 0)
            self.assertEqual(job["mask"]["LastEvent"], 99)

    def test50EventSplit(self):
        """
        _test50EventSplit_

        Test event based job splitting when the number of events per job is
        50, this should result in two jobs.
        """

        splitter = SplitterFactory()
        jobFactory = splitter(self.singleFileSubscription)

        eventsPerJob = 50
        jobGroups = jobFactory(events_per_job=eventsPerJob,
                               performance=self.performanceParams)

        assert len(jobGroups) == 1, \
            "ERROR: JobFactory didn't return one JobGroup."

        assert len(jobGroups[0].jobs) == 2, \
            "ERROR: JobFactory created %s jobs not two" % len(jobGroups[0].jobs)

        firstEvents = []
        for job in jobGroups[0].jobs:
            assert job.getFiles(type="lfn") == ["/some/file/name"], \
                "ERROR: Job contains unknown files."

            assert job["mask"].getMaxEvents() in [eventsPerJob, 1], \
                "ERROR: Job's max events is incorrect."

            assert job["mask"]["FirstEvent"] in [0, eventsPerJob], \
                "ERROR: Job's first event is incorrect."

            assert job["mask"]["FirstEvent"] not in firstEvents, \
                "ERROR: Job's first event is repeated."
            firstEvents.append(job["mask"]["FirstEvent"])

        return

    def test99EventSplit(self):
        """
        _test99EventSplit_

        Test event based job splitting when the number of events per job is
        99, this should result in two jobs.
        """
        splitter = SplitterFactory()
        jobFactory = splitter(self.singleFileSubscription)

        eventsPerJob = 99
        jobGroups = jobFactory(events_per_job=eventsPerJob,
                               performance=self.performanceParams)

        assert len(jobGroups) == 1, \
            "ERROR: JobFactory didn't return one JobGroup."

        assert len(jobGroups[0].jobs) == 2, \
            "ERROR: JobFactory created %s jobs not two" % len(jobGroups[0].jobs)

        firstEvents = []
        for job in jobGroups[0].jobs:
            assert job.getFiles(type="lfn") == ["/some/file/name"], \
                "ERROR: Job contains unknown files."
            self.assertTrue(job["mask"].getMaxEvents() in [eventsPerJob, 1],
                            "ERROR: Job's max events is incorrect.")

            assert job["mask"]["FirstEvent"] in [0, eventsPerJob], \
                "ERROR: Job's first event is incorrect."

            assert job["mask"]["FirstEvent"] not in firstEvents, \
                "ERROR: Job's first event is repeated."
            firstEvents.append(job["mask"]["FirstEvent"])

        return

    def test100EventMultipleFileSplit(self):
        """
        _test100EventMultipleFileSplit_

        Test job splitting into 100 event jobs when the input subscription has
        more than one file available.
        """
        splitter = SplitterFactory()
        jobFactory = splitter(self.multipleFileSubscription)

        jobGroups = jobFactory(events_per_job=self.eventsPerJob,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1, "ERROR: JobFactory didn't return one JobGroup.")

        self.assertEqual(len(jobGroups[0].jobs), 10,
                         "ERROR: JobFactory created %s jobs not ten" % len(jobGroups[0].jobs))

        for job in jobGroups[0].jobs:
            self.assertEqual(len(job.getFiles(type="lfn")), 1, "ERROR: Job contains too many files.")
            self.assertEqual(job["mask"].getMaxEvents(), self.eventsPerJob,
                             "ERROR: Job's max events is incorrect.")
            self.assertEqual(job["mask"]["FirstEvent"], 0, "ERROR: Job's first event is incorrect.")

        return

    def test50EventMultipleFileSplit(self):
        """
        _test50EventMultipleFileSplit_

        Test job splitting into 50 event jobs when the input subscription has
        more than one file available.
        """

        splitter = SplitterFactory()
        jobFactory = splitter(self.multipleFileSubscription)

        eventsPerJob = 50
        jobGroups = jobFactory(events_per_job=eventsPerJob,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1, "ERROR: JobFactory didn't return one JobGroup.")

        self.assertEqual(len(jobGroups[0].jobs), 20,
                         "ERROR: JobFactory created %s jobs not twenty" % len(jobGroups[0].jobs))

        for job in jobGroups[0].jobs:
            self.assertEqual(len(job.getFiles(type="lfn")), 1, "ERROR: Job contains too many files.")

            self.assertEqual(job["mask"].getMaxEvents(), eventsPerJob,
                             "ERROR: Job's max events is incorrect.")

            self.assertTrue(job["mask"]["FirstEvent"] in [0, eventsPerJob],
                            "ERROR: Job's first event is incorrect.")

        return

    def test150EventMultipleFileSplit(self):
        """
        _test150EventMultipleFileSplit_

        Test job splitting into 150 event jobs when the input subscription has
        more than one file available.  This test verifies that the job splitting
        code will put at most one file in a job.
        """
        splitter = SplitterFactory()
        jobFactory = splitter(self.multipleFileSubscription)

        jobGroups = jobFactory(events_per_job=150,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1)

        self.assertEqual(len(jobGroups[0].jobs), 10)

        for job in jobGroups[0].jobs:
            self.assertEqual(len(job.getFiles(type="lfn")), 1)
            self.assertEqual(job["mask"].getMaxEvents(), self.eventsPerJob)
            self.assertEqual(job["mask"]["FirstEvent"], 0)
            self.assertEqual(job["mask"]["LastEvent"], 99)

    def testMCExactEvents(self):
        """
        _testMCExactEvents_
        Test event based job splitting when the number of events per job is
        exactly the same as the number of events in the input file and no lumi
        information was supplied.
        """
        singleMCSubscription = self.generateFakeMCFile(firstLumi=1,
                                                       lastLumi=1)
        splitter = SplitterFactory()
        jobFactory = splitter(singleMCSubscription)

        jobGroups = jobFactory(events_per_job=self.eventsPerJob,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1,
                         "Error: JobFactory did not return one JobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 1,
                         "Error: JobFactory created %s jobs not one"
                         % len(jobGroups[0].jobs))

        job = jobGroups[0].jobs.pop()

        self.assertEqual(job.getFiles(type="lfn"), ["MCFakeFileTest"],
                         "Error: Job contains unknown files.")

        self.assertEqual(job["mask"].getMaxEvents(), 100,
                         "Error: Job's max events is incorrect.%i" % job["mask"].getMaxEvents())
        self.assertEqual(job["mask"]["FirstEvent"], 1,
                         "Error: Job's first event is incorrect.")
        self.assertEqual(job["mask"]["FirstLumi"], 1,
                         "Error: Job's first lumi is incorrect.")
        self.assertEqual(len(job["mask"].getRunAndLumis()), 0,
                         "Error: Job's mask has runs and lumis")

    def testMCMoreEvents(self):
        """
        _testMCMoreEvents_

        Test event based job splitting when the number of events per job is
        greater than the number of events in the input file and no lumi
        information was supplied.
        """
        singleMCSubscription = self.generateFakeMCFile(firstLumi=1,
                                                       lastLumi=1)
        splitter = SplitterFactory()
        jobFactory = splitter(singleMCSubscription)

        jobGroups = jobFactory(events_per_job=1000,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1,
                         "Error: JobFactory did not return one JobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 1,
                         "Error: JobFactory created %s jobs not one"
                         % len(jobGroups[0].jobs))

        job = jobGroups[0].jobs.pop()

        self.assertEqual(job.getFiles(type="lfn"), ["MCFakeFileTest"],
                         "Error: Job contains unknown files.")

        self.assertEqual(job["mask"].getMaxEvents(), 100,
                         "Error: Job's max events is incorrect.")
        self.assertEqual(job["mask"]["FirstEvent"], 1,
                         "Error: Job's first event is incorrect.")
        self.assertEqual(job["mask"]["FirstLumi"], 1,
                         "Error: Job's first lumi is incorrect.")
        self.assertEqual(len(job["mask"].getRunAndLumis()), 0,
                         "Error: Job's mask has runs and lumis")

    def testMC99EventSplit(self):
        """
        _testMC99EventSplit_

        Test event based job splitting when the number of events per job is
        99, this should result in two jobs.
        No lumi information is supplied here.
        """
        singleMCSubscription = self.generateFakeMCFile(firstLumi=1,
                                                       lastLumi=2)
        splitter = SplitterFactory()
        jobFactory = splitter(singleMCSubscription)

        jobGroups = jobFactory(events_per_job=99,
                               lheInputFiles=True,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1,
                         "Error: JobFactory did not return one JobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 2,
                         "Error: JobFactory created %s jobs not two"
                         % len(jobGroups[0].jobs))
        for job in jobGroups[0].jobs:
            firstJobCondition = (job["mask"].getMaxEvents() == 99 and
                                 job["mask"]["FirstLumi"] == 1 and
                                 job["mask"]["FirstEvent"] == 1)
            secondJobCondition = (job["mask"].getMaxEvents() == 1 and
                                  job["mask"]["FirstLumi"] == 2 and
                                  job["mask"]["FirstEvent"] == 100)
            self.assertTrue(firstJobCondition or secondJobCondition,
                            "Job mask: %s didn't pass neither of the conditions"
                            % job["mask"])
            self.assertTrue(job.getBaggage().lheInputFiles)

    def testMC50EventSplit(self):
        """
        _testMC50EventSplit_

        Test event based job splitting when the number of events per job is
        50, this should result in two jobs.
        No lumi information supplied here.
        """
        singleMCSubscription = self.generateFakeMCFile(firstLumi=1,
                                                       lastLumi=2)
        splitter = SplitterFactory()
        jobFactory = splitter(singleMCSubscription)

        jobGroups = jobFactory(events_per_job=50,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1,
                         "Error: JobFactory did not return one JobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 2,
                         "Error: JobFactory created %s jobs not two"
                         % len(jobGroups[0].jobs))
        for job in jobGroups[0].jobs:
            firstJobCondition = (job["mask"].getMaxEvents() == 50 and
                                 job["mask"]["FirstLumi"] == 1 and
                                 job["mask"]["FirstEvent"] == 1)
            secondJobCondition = (job["mask"].getMaxEvents() == 50 and
                                  job["mask"]["FirstLumi"] == 2 and
                                  job["mask"]["FirstEvent"] == 51)
            self.assertTrue(firstJobCondition or secondJobCondition,
                            "Job mask: %s didn't pass neither of the conditions"
                            % job["mask"])
            self.assertFalse(job.getBaggage().lheInputFiles)

        return

    def testMCShiftedEventSplit(self):
        """
        _testMCShiftedEventSplit_

        Performs different tests with files that start with event counters
        different than 1, lumi information remains default.
        """
        singleMCSubscription = self.generateFakeMCFile(numEvents=600,
                                                       firstEvent=201,
                                                       lastEvent=800)
        splitter = SplitterFactory()
        jobFactory = splitter(singleMCSubscription)

        jobGroups = jobFactory(events_per_job=600,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1,
                         "Error: JobFactory did not return one JobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 1,
                         "Error: JobFactory created %s jobs not one"
                         % len(jobGroups[0].jobs))
        job = jobGroups[0].jobs.pop()
        self.assertEqual(job["mask"].getMaxEvents(), 600,
                         "Error: Job's max events is incorrect.")
        self.assertEqual(job["mask"]["FirstEvent"], 201,
                         "Error: Job's first event is incorrect.")
        self.assertEqual(job["mask"]["FirstLumi"], 1,
                         "Error: Job's first lumi is incorrect.")

        singleMCSubscription = self.generateFakeMCFile(numEvents=600,
                                                       firstEvent=201,
                                                       lastEvent=800)
        splitter = SplitterFactory()
        jobFactory = splitter(singleMCSubscription)
        jobGroups = jobFactory(events_per_job=6000,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1,
                         "Error: JobFactory did not return one JobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 1,
                         "Error: JobFactory created %s jobs not one"
                         % len(jobGroups[0].jobs))
        job = jobGroups[0].jobs.pop()
        self.assertEqual(job["mask"].getMaxEvents(), 600,
                         "Error: Job's max events is incorrect.")
        self.assertEqual(job["mask"]["FirstEvent"], 201,
                         "Error: Job's first event is incorrect.")
        self.assertEqual(job["mask"]["FirstLumi"], 1,
                         "Error: Job's first lumi is incorrect.")

        singleMCSubscription = self.generateFakeMCFile(numEvents=600,
                                                       firstEvent=201,
                                                       lastEvent=800)
        splitter = SplitterFactory()
        jobFactory = splitter(singleMCSubscription)
        jobGroups = jobFactory(events_per_job=599,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1,
                         "Error: JobFactory did not return one JobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 2,
                         "Error: JobFactory created %s jobs not two"
                         % len(jobGroups[0].jobs))
        for job in jobGroups[0].jobs:
            firstJobCondition = (job["mask"].getMaxEvents() == 599 and
                                 job["mask"]["FirstLumi"] == 1 and
                                 job["mask"]["FirstEvent"] == 201)
            secondJobCondition = (job["mask"].getMaxEvents() == 1 and
                                  job["mask"]["FirstLumi"] == 2 and
                                  job["mask"]["FirstEvent"] == 800)
            self.assertTrue(firstJobCondition or secondJobCondition,
                            "Job mask: %s didn't pass neither of the conditions"
                            % job["mask"])

        singleMCSubscription = self.generateFakeMCFile(numEvents=600,
                                                       firstEvent=201,
                                                       lastEvent=800)
        splitter = SplitterFactory()
        jobFactory = splitter(singleMCSubscription)
        jobGroups = jobFactory(events_per_job=300,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1,
                         "Error: JobFactory did not return one JobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 2,
                         "Error: JobFactory created %s jobs not two"
                         % len(jobGroups[0].jobs))
        for job in jobGroups[0].jobs:
            firstJobCondition = (job["mask"].getMaxEvents() == 300 and
                                 job["mask"]["FirstLumi"] == 1 and
                                 job["mask"]["FirstEvent"] == 201)
            secondJobCondition = (job["mask"].getMaxEvents() == 300 and
                                  job["mask"]["FirstLumi"] == 2 and
                                  job["mask"]["FirstEvent"] == 501)
            self.assertTrue(firstJobCondition or secondJobCondition,
                            "Job mask: %s didn't pass neither of the conditions"
                            % job["mask"])

    def testMCShiftedLumiSplit(self):
        """
        _testMCShiftedLumiSplit

        Perform different tests with files that have lumi counters starting
        in something different than 1, however the splitting algorithm
        splits lumi with it's default value.
        """
        singleMCSubscription = self.generateFakeMCFile(firstLumi=345,
                                                       lastLumi=345)
        splitter = SplitterFactory()
        jobFactory = splitter(singleMCSubscription)

        jobGroups = jobFactory(events_per_job=100,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1,
                         "Error: JobFactory did not return one JobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 1,
                         "Error: JobFactory created %s jobs not one"
                         % len(jobGroups[0].jobs))
        job = jobGroups[0].jobs.pop()
        self.assertEqual(job["mask"].getMaxEvents(), 100,
                         "Error: Job's max events is incorrect.")
        self.assertEqual(job["mask"]["FirstEvent"], 1,
                         "Error: Job's first event is incorrect.")
        self.assertEqual(job["mask"]["FirstLumi"], 345,
                         "Error: Job's first lumi is incorrect.")

        singleMCSubscription = self.generateFakeMCFile(firstLumi=345,
                                                       lastLumi=345)
        splitter = SplitterFactory()
        jobFactory = splitter(singleMCSubscription)
        jobGroups = jobFactory(events_per_job=1000,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1,
                         "Error: JobFactory did not return one JobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 1,
                         "Error: JobFactory created %s jobs not one"
                         % len(jobGroups[0].jobs))
        job = jobGroups[0].jobs.pop()
        self.assertEqual(job["mask"].getMaxEvents(), 100,
                         "Error: Job's max events is incorrect.")
        self.assertEqual(job["mask"]["FirstEvent"], 1,
                         "Error: Job's first event is incorrect.")
        self.assertEqual(job["mask"]["FirstLumi"], 345,
                         "Error: Job's first lumi is incorrect.")

        singleMCSubscription = self.generateFakeMCFile(firstLumi=345,
                                                       lastLumi=345)
        splitter = SplitterFactory()
        jobFactory = splitter(singleMCSubscription)
        jobGroups = jobFactory(events_per_job=99,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1,
                         "Error: JobFactory did not return one JobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 2,
                         "Error: JobFactory created %s jobs not two"
                         % len(jobGroups[0].jobs))
        for job in jobGroups[0].jobs:
            firstJobCondition = (job["mask"].getMaxEvents() == 99 and
                                 job["mask"]["FirstLumi"] == 345 and
                                 job["mask"]["FirstEvent"] == 1)
            secondJobCondition = (job["mask"].getMaxEvents() == 1 and
                                  job["mask"]["FirstLumi"] == 346 and
                                  job["mask"]["FirstEvent"] == 100)
            self.assertTrue(firstJobCondition or secondJobCondition,
                            "Job mask: %s didn't pass neither of the conditions"
                            % job["mask"])

        singleMCSubscription = self.generateFakeMCFile(firstLumi=345,
                                                       lastLumi=345)
        splitter = SplitterFactory()
        jobFactory = splitter(singleMCSubscription)
        jobGroups = jobFactory(events_per_job=50,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1,
                         "Error: JobFactory did not return one JobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 2,
                         "Error: JobFactory created %s jobs not two"
                         % len(jobGroups[0].jobs))
        for job in jobGroups[0].jobs:
            firstJobCondition = (job["mask"].getMaxEvents() == 50 and
                                 job["mask"]["FirstLumi"] == 345 and
                                 job["mask"]["FirstEvent"] == 1)
            secondJobCondition = (job["mask"].getMaxEvents() == 50 and
                                  job["mask"]["FirstLumi"] == 346 and
                                  job["mask"]["FirstEvent"] == 51)
            self.assertTrue(firstJobCondition or secondJobCondition,
                            "Job mask: %s didn't pass neither of the conditions"
                            % job["mask"])

    def testMCLumiSplit(self):
        """
        _testMCLumiSplit_

        2 tests on lumi splitting are performed:
            1. The number of events per job is a multiple of the events
            per lumi
            2. The number of events per job is not a multiple of the events
            per lumi

        """
        singleMCSubscription = self.generateFakeMCFile(numEvents=150,
                                                       lastEvent=150,
                                                       lastLumi=15)
        splitter = SplitterFactory()
        jobFactory = splitter(singleMCSubscription)

        jobGroups = jobFactory(events_per_job=100, events_per_lumi=10,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1,
                         "Error: JobFactory did not return one JobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 2,
                         "Error: JobFactory created %s jobs not two"
                         % len(jobGroups[0].jobs))
        for job in jobGroups[0].jobs:
            firstJobCondition = (job["mask"].getMaxEvents() == 100 and
                                 job["mask"]["FirstLumi"] == 1 and job["mask"]["LastLumi"] == 10 and
                                 job["mask"]["FirstRun"] == 1 and job["mask"]["LastRun"] == 1 and
                                 job["mask"]["FirstEvent"] == 1 and job["mask"]["LastEvent"] == 100)
            secondJobCondition = (job["mask"].getMaxEvents() == 50 and
                                  job["mask"]["FirstLumi"] == 11 and job["mask"]["LastLumi"] == 15 and
                                  job["mask"]["FirstRun"] == 1 and job["mask"]["LastRun"] == 1 and
                                  job["mask"]["FirstEvent"] == 101 and job["mask"]["LastEvent"] == 150)
            self.assertTrue(firstJobCondition or secondJobCondition,
                            "Job mask: %s didn't pass neither of the conditions" % job["mask"])

        singleMCSubscription = self.generateFakeMCFile(numEvents=150,
                                                       lastEvent=150,
                                                       lastLumi=15)
        splitter = SplitterFactory()
        jobFactory = splitter(singleMCSubscription)

        jobGroups = jobFactory(events_per_job=111, events_per_lumi=10,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1,
                         "Error: JobFactory did not return one JobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 2,
                         "Error: JobFactory created %s jobs not two"
                         % len(jobGroups[0].jobs))
        for job in jobGroups[0].jobs:
            firstJobCondition = (job["mask"].getMaxEvents() == 111 and
                                 job["mask"]["FirstLumi"] == 1 and job["mask"]["LastLumi"] == 12 and
                                 job["mask"]["FirstRun"] == 1 and job["mask"]["LastRun"] == 1 and
                                 job["mask"]["FirstEvent"] == 1 and job["mask"]["LastEvent"] == 111)
            secondJobCondition = (job["mask"].getMaxEvents() == 39 and
                                  job["mask"]["FirstLumi"] == 13 and job["mask"]["LastLumi"] == 16 and
                                  job["mask"]["FirstRun"] == 1 and job["mask"]["LastRun"] == 1 and
                                  job["mask"]["FirstEvent"] == 112 and job["mask"]["LastEvent"] == 150)
            self.assertTrue(firstJobCondition or secondJobCondition,
                            "Job mask: %s didn't pass neither of the conditions" % job["mask"])

    def testMCEventSplitOver32bit(self):
        """
        _testMCEventSplitOver32bit_

        Make sure that no events will go over a 32 bit unsigned integer
        representation, event counter should be reset in that case.
        Also test is not over cautious.
        """
        firstEvent = 1
        singleMCSubscription = self.generateFakeMCFile(numEvents=2 ** 32,
                                                       firstEvent=firstEvent)
        splitter = SplitterFactory()
        jobFactory = splitter(singleMCSubscription)

        jobGroups = jobFactory(events_per_job=2 ** 32 - 1,
                               events_per_lumi=2 ** 32 - 1,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1,
                         "Error: JobFactory did not return one JobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 2,
                         "Error: JobFactory created %s jobs not two"
                         % len(jobGroups[0].jobs))
        for job in jobGroups[0].jobs:
            firstJobCondition = (job["mask"].getMaxEvents() == 2 ** 32 - 1 and
                                 job["mask"]["FirstLumi"] == 1 and
                                 job["mask"]["FirstEvent"] == firstEvent and
                                 job["mask"]["LastEvent"] <= 2 ** 32)
            secondJobCondition = (job["mask"].getMaxEvents() == 1 and
                                  job["mask"]["FirstLumi"] == 2 and
                                  job["mask"]["FirstEvent"] == 1)
            self.assertTrue(firstJobCondition or secondJobCondition,
                            "Job mask: %s didn't pass neither of the conditions"
                            % job["mask"])

        # File not exceeding the unsigned 32 bits
        singleMCSubscription = self.generateFakeMCFile(numEvents=2 ** 32 - 2,
                                                       firstEvent=firstEvent)
        jobFactory = splitter(singleMCSubscription)

        jobGroups = jobFactory(events_per_job=2 ** 31, events_per_lumi=2 ** 32,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1,
                         "Error: JobFactory did not return one JobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 2,
                         "Error: JobFactory created %s jobs not two"
                         % len(jobGroups[0].jobs))
        for job in jobGroups[0].jobs:
            firstJobCondition = (job["mask"].getMaxEvents() == 2 ** 31 and
                                 job["mask"]["FirstLumi"] == 1 and
                                 job["mask"]["FirstEvent"] == firstEvent and
                                 job["mask"]["LastEvent"] <= 2 ** 31
                                 )
            secondJobCondition = (job["mask"].getMaxEvents() == 2 ** 31 - 2 and
                                  job["mask"]["FirstLumi"] == 2 and
                                  job["mask"]["FirstEvent"] == 2 ** 31 + 1 and
                                  job["mask"]["LastEvent"] <= 2 ** 32 - 2)
            self.assertTrue(firstJobCondition or secondJobCondition,
                            "Job mask: %s didn't pass neither of the conditions"
                            % job["mask"])
        firstEvent = 1
        singleMCSubscription = self.generateFakeMCFile(numEvents=2 ** 32 - 1,
                                                       firstEvent=firstEvent)
        splitter = SplitterFactory()
        jobFactory = splitter(singleMCSubscription)

        jobGroups = jobFactory(events_per_job=2 ** 32, events_per_lumi=2 ** 32,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1,
                         "Error: JobFactory did not return one JobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 1,
                         "Error: JobFactory created %s jobs not one"
                         % len(jobGroups[0].jobs))

        job = jobGroups[0].jobs.pop()

        self.assertEqual(job["mask"].getMaxEvents(), 2 ** 32 - 1,
                         "Error: Job's max events is incorrect.")
        self.assertEqual(job["mask"]["FirstEvent"], 1,
                         "Error: Job's first event is incorrect.")
        self.assertEqual(job["mask"]["FirstLumi"], 1,
                         "Error: Job's first lumi is incorrect.")

        firstEvent = 2 ** 32 - 1
        singleMCSubscription = self.generateFakeMCFile(numEvents=2,
                                                       firstEvent=firstEvent)
        splitter = SplitterFactory()
        jobFactory = splitter(singleMCSubscription)

        jobGroups = jobFactory(events_per_job=3, events_per_lumi=1,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1,
                         "Error: JobFactory did not return one JobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 1,
                         "Error: JobFactory created %s jobs not one"
                         % len(jobGroups[0].jobs))

        job = jobGroups[0].jobs.pop()

        self.assertEqual(job["mask"].getMaxEvents(), 2,
                         "Error: Job's max events is incorrect.")
        self.assertEqual(job["mask"]["FirstEvent"], 1,
                         "Error: Job's first event is incorrect.")
        self.assertEqual(job["mask"]["FirstLumi"], 1,
                         "Error: Job's first lumi is incorrect.")

        firstEvent = 2 ** 32
        singleMCSubscription = self.generateFakeMCFile(numEvents=50,
                                                       firstEvent=firstEvent)
        splitter = SplitterFactory()
        jobFactory = splitter(singleMCSubscription)

        jobGroups = jobFactory(events_per_job=60, events_per_lumi=10,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1,
                         "Error: JobFactory did not return one JobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 1,
                         "Error: JobFactory created %s jobs not one"
                         % len(jobGroups[0].jobs))

        job = jobGroups[0].jobs.pop()

        self.assertEqual(job["mask"].getMaxEvents(), 50,
                         "Error: Job's max events is incorrect.")
        self.assertEqual(job["mask"]["FirstEvent"], 1,
                         "Error: Job's first event is incorrect.")
        self.assertEqual(job["mask"]["FirstLumi"], 1,
                         "Error: Job's first lumi is incorrect.")

        firstEvent = 2 ** 32
        singleMCSubscription = self.generateFakeMCFile(numEvents=50,
                                                       firstEvent=firstEvent)
        splitter = SplitterFactory()
        jobFactory = splitter(singleMCSubscription)

        jobGroups = jobFactory(events_per_job=30, events_per_lumi=10,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1,
                         "Error: JobFactory did not return one JobGroup")
        self.assertEqual(len(jobGroups[0].jobs), 2,
                         "Error: JobFactory created %s jobs not two"
                         % len(jobGroups[0].jobs))

        for job in jobGroups[0].jobs:
            firstJobCondition = (job["mask"].getMaxEvents() == 30 and
                                 job["mask"]["FirstLumi"] == 1 and
                                 job["mask"]["FirstEvent"] == 1 and
                                 job["mask"]["LastEvent"] <= 2 ** 32)
            secondJobCondition = (job["mask"].getMaxEvents() == 20 and
                                  job["mask"]["FirstLumi"] == 4 and
                                  job["mask"]["FirstEvent"] == 31)
            self.assertTrue(firstJobCondition or secondJobCondition,
                            "Job mask: %s didn't pass neither of the conditions"
                            % job["mask"])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_GetSiteSE_

Oracle implementation of Locations.GetSiteSE
"""

__all__ = []



from WMCore.WMBS.MySQL.Locations.GetSiteSE import GetSiteSE as MySQLGetSiteSE

class GetSiteSE(MySQLGetSiteSE):
    """
    Same as MySQL version

    """



#!/usr/bin/env python
"""
_JobGroup_t_

Unit tests for the WMBS JobGroup class.
"""

from builtins import range
import threading
import unittest
from pprint import pformat as pretty

from WMCore.DAOFactory import DAOFactory
from WMCore.DataStructs.Run import Run
from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset as WMBSFileset
from WMCore.WMBS.Job import Job
from WMCore.WMBS.JobGroup import JobGroup
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow
from WMQuality.TestInit import TestInit


class JobGroupTest(unittest.TestCase):
    def setUp(self):
        """
        _setUp_

        Setup the database and logging connection.  Try to create all of the
        WMBS tables.
        """

        myThread = threading.currentThread()

        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        self.testInit.setSchema(customModules=["WMCore.WMBS"],
                                useDefault=False)

        # We need to set sites in the locations table
        daofactory = DAOFactory(package="WMCore.WMBS",
                                logger=myThread.logger,
                                dbinterface=myThread.dbi)

        locationAction = daofactory(classname="Locations.New")
        locationAction.execute(siteName="site1", pnn="T2_CH_CERN")
        locationAction.execute(siteName="site2", pnn="malpaquet")
        locationAction.execute(siteName="site3", pnn="badse.cern.ch")

    def tearDown(self):
        """
        _tearDown_

        Drop all the WMBS tables.
        """
        self.testInit.clearDatabase()

    def createTestJobGroup(self, commitFlag=True):
        """
        _createTestJobGroup_

        """
        testWorkflow = Workflow(spec="spec.xml", owner="Simon",
                                name="wf001", task="Test")
        testWorkflow.create()

        testWMBSFileset = WMBSFileset(name="TestFileset")
        testWMBSFileset.create()

        testSubscription = Subscription(fileset=testWMBSFileset,
                                        workflow=testWorkflow)
        testSubscription.create()

        testJobGroup = JobGroup(subscription=testSubscription)
        testJobGroup.create()

        testFileA = File(lfn="/this/is/a/lfnA", size=1024, events=10)
        testFileA.addRun(Run(10, *[12312]))
        testFileA.setLocation("T2_CH_CERN")
        testFileA.setLocation("malpaquet")

        testFileB = File(lfn="/this/is/a/lfnB", size=1024, events=10)
        testFileB.addRun(Run(10, *[12312]))
        testFileB.setLocation("T2_CH_CERN")
        testFileB.setLocation("malpaquet")

        testFileA.create()
        testFileB.create()

        testJobA = Job(name="TestJobA")
        testJobA.addFile(testFileA)

        testJobB = Job(name="TestJobB")
        testJobB.addFile(testFileB)

        testJobGroup.add(testJobA)
        testJobGroup.add(testJobB)

        if commitFlag:
            testJobGroup.commit()

        return testJobGroup

    def createLargerTestJobGroup(self, commitFlag=True):
        """
        _createTestJobGroup_

        """
        testWorkflow = Workflow(spec="spec.xml", owner="Simon",
                                name="wf002", task="Test2")
        testWorkflow.create()

        testWMBSFileset = WMBSFileset(name="TestFileset")
        testWMBSFileset.create()

        testSubscription = Subscription(fileset=testWMBSFileset,
                                        workflow=testWorkflow)
        testSubscription.create()

        testJobGroup = JobGroup(subscription=testSubscription)
        testJobGroup.create()

        testFileC = File(lfn="/this/is/a/lfnC", size=1024, events=10)
        testFileC.addRun(Run(10, *[12312]))
        testFileC.setLocation("T2_CH_CERN")
        testFileC.setLocation("malpaquet")

        testFileD = File(lfn="/this/is/a/lfnD", size=1024, events=10)
        testFileD.addRun(Run(10, *[12312]))
        testFileD.setLocation("T2_CH_CERN")
        testFileD.setLocation("malpaquet")

        testFileC.create()
        testFileD.create()

        testJobA = Job(name="TestJobA1")
        testJobA.addFile(testFileC)

        testJobB = Job(name="TestJobB1")
        testJobB.addFile(testFileD)

        testJobGroup.add(testJobA)
        testJobGroup.add(testJobB)

        for i in range(0, 100):
            testJob = Job(name="TestJob%i" % (i))
            testJob.addFile(testFileC)
            testJobGroup.add(testJob)

        if commitFlag:
            testJobGroup.commit()

        return testJobGroup

    def testCreateDeleteExists(self):
        """
        _testCreateDeleteExists_

        Create a JobGroup and then delete it.  Use the JobGroup's exists()
        method to determine if it exists before it is created, after it is
        created and after it is deleted.
        """
        testWorkflow = Workflow(spec="spec.xml", owner="Simon",
                                name="wf001", task="Test")
        testWorkflow.create()

        testFileset = WMBSFileset(name="TestFileset")
        testFileset.create()

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=testWorkflow)
        testSubscription.create()

        testJobGroup = JobGroup(subscription=testSubscription)

        self.assertFalse(testJobGroup.exists())

        testJobGroup.create()

        self.assertTrue(testJobGroup.exists())

        testJobGroup.delete()

        self.assertFalse(testJobGroup.exists())

        testSubscription.delete()
        testFileset.delete()
        testWorkflow.delete()
        return

    def testCreateTransaction(self):
        """
        _testCreateTransaction_

        Create a JobGroup and commit it to the database.  Rollback the database
        transaction and verify that the JobGroup is no longer in the database.
        """
        testWorkflow = Workflow(spec="spec.xml", owner="Simon",
                                name="wf001", task="Test")
        testWorkflow.create()

        testFileset = WMBSFileset(name="TestFileset")
        testFileset.create()

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=testWorkflow)
        testSubscription.create()

        testJobGroup = JobGroup(subscription=testSubscription)

        self.assertFalse(testJobGroup.exists())

        myThread = threading.currentThread()
        myThread.transaction.begin()

        testJobGroup.create()

        self.assertTrue(testJobGroup.exists())

        myThread.transaction.rollback()

        self.assertFalse(testJobGroup.exists())

        testSubscription.delete()
        testFileset.delete()
        testWorkflow.delete()
        return

    def testDeleteTransaction(self):
        """
        _testDeleteTransaction_

        Create a JobGroup and then commit it to the database.  Begin a
        transaction and the delete the JobGroup from the database.  Using the
        exists() method verify that the JobGroup is not in the database.
        Finally, roll back the transaction and verify that the JobGroup is
        in the database.
        """
        testWorkflow = Workflow(spec="spec.xml", owner="Simon",
                                name="wf001", task="Test")
        testWorkflow.create()

        testFileset = WMBSFileset(name="TestFileset")
        testFileset.create()

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=testWorkflow)
        testSubscription.create()

        testJobGroup = JobGroup(subscription=testSubscription)

        self.assertFalse(testJobGroup.exists())

        testJobGroup.create()

        self.assertTrue(testJobGroup.exists())

        myThread = threading.currentThread()
        myThread.transaction.begin()

        testJobGroup.delete()

        self.assertFalse(testJobGroup.exists())

        myThread.transaction.rollback()

        self.assertTrue(testJobGroup.exists())

        testSubscription.delete()
        testFileset.delete()
        testWorkflow.delete()
        return

    def testLoad(self):
        """
        _testLoad_

        Test loading the JobGroup and any associated meta data from the
        database.
        """
        testJobGroupA = self.createTestJobGroup()

        testJobGroupB = JobGroup(id=testJobGroupA.id)
        testJobGroupB.load()
        testJobGroupC = JobGroup(uid=testJobGroupA.uid)
        testJobGroupC.load()

        self.assertIsInstance(testJobGroupB.id, int)
        self.assertIsInstance(testJobGroupC.id, int)
        self.assertIsInstance(testJobGroupB.subscription["id"], int)
        self.assertIsInstance(testJobGroupC.subscription["id"], int)
        self.assertIsInstance(testJobGroupB.output.id, int)
        self.assertIsInstance(testJobGroupC.output.id, int)

        assert testJobGroupB.uid == testJobGroupA.uid, \
            "ERROR: Job group did not load uid correctly."

        assert testJobGroupC.id == testJobGroupA.id, \
            "ERROR: Job group did not load id correctly."

        assert testJobGroupB.subscription["id"] == \
               testJobGroupA.subscription["id"], \
            "ERROR: Job group did not load subscription correctly"

        assert testJobGroupC.subscription["id"] == \
               testJobGroupA.subscription["id"], \
            "ERROR: Job group did not load subscription correctly"

        assert testJobGroupB.output.id == testJobGroupA.output.id, \
            "ERROR: Output fileset didn't load properly"

        assert testJobGroupC.output.id == testJobGroupA.output.id, \
            "ERROR: Output fileset didn't load properly"

        return

    def testLoadData(self):
        """
        _testLoadData_

        Test loading the JobGroup, it's meta data and any data associated with
        its output fileset and jobs from the database.
        """
        testJobGroupA = self.createTestJobGroup()

        testJobGroupB = JobGroup(id=testJobGroupA.id)
        testJobGroupB.loadData()

        self.assertEqual(testJobGroupB.subscription["id"], testJobGroupA.subscription["id"],
                         "Job group did not load subscription correctly")

        # Build a tuple with important job information for each job group
        goldenJobs = [(job['id'], job['name'],
                       sorted([inp_file['lfn'] for inp_file in job['input_files']]))
                      for job in testJobGroupA.getJobs(type="list")]
        loadedJobs = [(job['id'], job['name'],
                       sorted([inp_file['lfn'] for inp_file in job['input_files']]))
                      for job in testJobGroupB.getJobs(type="list")]

        # Make sure each job from one group is in the other list
        for job in loadedJobs:
            self.assertIn(job, goldenJobs, "JobGroup loaded an unknown job: \n%s \nis not in  \n%s" %
                          (pretty(job), pretty(goldenJobs)))
            goldenJobs.remove(job)

        self.assertFalse(goldenJobs, "JobGroup didn't load all jobs")
        self.assertEqual(testJobGroupB.output.id, testJobGroupA.output.id, "Output fileset didn't load properly")

        return

    def testCommit(self):
        """
        _testCommit_

        Verify that jobs are not added to a job group until commit() is called
        on the JobGroup.  Also verify that commit() correctly commits the jobs
        to the database.
        """
        testJobGroupA = self.createTestJobGroup(commitFlag=False)

        testJobGroupB = JobGroup(id=testJobGroupA.id)
        testJobGroupB.loadData()

        assert len(testJobGroupA.getJobs()) == 0, \
            "ERROR: Original object commited too early"
        assert len(testJobGroupB.getJobs()) == 0, \
            "ERROR: Loaded JobGroup has too many jobs"

        testJobGroupA.commit()
        testJobGroupA.loadData()

        assert len(testJobGroupA.getJobs()) == 2, \
            "ERROR: Original object did not commit jobs"

        testJobGroupC = JobGroup(id=testJobGroupA.id)
        testJobGroupC.loadData()

        assert len(testJobGroupC.getJobs()) == 2, \
            "ERROR: Loaded object has too few jobs."

        return

    def testCommitTransaction(self):
        """
        _testCommitTransaction_

        Create a JobGroup and then add some jobs to it.  Begin a transaction
        and then call commit() on the JobGroup.  Verify that the newly committed
        jobs can be loaded from the database.  Rollback the transaction and then
        verify that the jobs that were committed before are no longer associated
        with the JobGroup.
        """
        testJobGroupA = self.createTestJobGroup(commitFlag=False)

        testJobGroupB = JobGroup(id=testJobGroupA.id)
        testJobGroupB.loadData()

        assert len(testJobGroupA.getJobs()) == 0, \
            "ERROR: Original object commited too early"

        assert len(testJobGroupB.getJobs()) == 0, \
            "ERROR: Loaded JobGroup has too many jobs"

        myThread = threading.currentThread()
        myThread.transaction.begin()

        testJobGroupA.commit()

        assert len(testJobGroupA.getJobs()) == 2, \
            "ERROR: Original object did not commit jobs"

        testJobGroupC = JobGroup(id=testJobGroupA.id)
        testJobGroupC.loadData()

        assert len(testJobGroupC.getJobs()) == 2, \
            "ERROR: Loaded object has too few jobs."

        myThread.transaction.rollback()

        testJobGroupD = JobGroup(id=testJobGroupA.id)
        testJobGroupD.loadData()

        assert len(testJobGroupD.getJobs()) == 0, \
            "ERROR: Loaded object has too many jobs."

        return

    def testSetGetSite(self):
        """
        _testSetGetSite_

        For the JobCreator we have to be able to specify the site for a JobGroup.
        This function tests the accessors for that information.
        """

        testJobGroup = self.createTestJobGroup()
        testJobGroup.setSite("site1")
        result = testJobGroup.getSite()

        self.assertEqual(result, "site1")

        return

    def testCommitBulk(self):
        """
        _testCommitBulk_

        Exactly the same as testCommit, but using commitBulk() instead of commit()
        """

        testJobGroupA = self.createLargerTestJobGroup(commitFlag=False)

        testJobGroupB = JobGroup(id=testJobGroupA.id)
        testJobGroupB.loadData()

        assert len(testJobGroupA.getJobs()) == 0, \
            "ERROR: Original object commited too early"
        assert len(testJobGroupB.getJobs()) == 0, \
            "ERROR: Loaded JobGroup has too many jobs"

        testJobGroupA.commitBulk()
        testJobGroupA.loadData()

        self.assertEqual(len(testJobGroupA.getJobs()), 102)

        testJobGroupC = JobGroup(id=testJobGroupA.id)
        testJobGroupC.loadData()

        self.assertEqual(len(testJobGroupC.getJobs()), 102)

        self.assertEqual(testJobGroupC.jobs[0].getFiles()[0]['lfn'], '/this/is/a/lfnC')
        self.assertEqual(testJobGroupC.jobs[1].getFiles()[0]['lfn'], '/this/is/a/lfnD')

        return

    def notestTwoJobGroups(self):
        """
        Test two job groups with a shared fileset. (Minimal part of testGetLocations which was failing)
        """

        testWorkflow1 = Workflow(spec="spec.xml", owner="Simon", name="wf001", task="Test1")
        testWorkflow1.create()

        testWMBSFileset1 = WMBSFileset(name="TestFileset1")
        testWMBSFileset1.create()

        testSubscription1 = Subscription(fileset=testWMBSFileset1, workflow=testWorkflow1)
        testSubscription1.create()

        testJobGroup1 = JobGroup(subscription=testSubscription1)
        testJobGroup1.create()

        testFileA = File(lfn="/this/is/a/lfnA", size=1024, events=10)
        testFileA.addRun(Run(10, *[12312]))
        testFileA.create()

        testJobA = Job(name="TestJobA")
        testJobA.addFile(testFileA)
        testJobGroup1.add(testJobA)

        testJobGroup1.commit()

        testWorkflow2 = Workflow(spec="spec.xml", owner="Simon", name="wf002", task="Test2")
        testWorkflow2.create()

        testWMBSFileset2 = WMBSFileset(name="TestFileset1")
        testWMBSFileset2.create()

        testSubscription2 = Subscription(fileset=testWMBSFileset2, workflow=testWorkflow2)
        testSubscription2.create()

        testJobGroup2 = JobGroup(subscription=testSubscription2)
        testJobGroup2.create()

        testFileC = File(lfn="/this/is/a/lfnC", size=1024, events=10)
        testFileC.addRun(Run(10, *[12312]))

        testFileC.create()

        testJobA1 = Job(name="TestJobA1")
        testJobA1.addFile(testFileC)

        testJobGroup2.add(testJobA1)
        testJobGroup2.commit()

    def testGetLocationsForJobs(self):
        """
        _testGetLocationsForJobs

        Tests the functionality of grabbing locations for a single job

        """

        testJobGroup = self.createTestJobGroup()

        result = testJobGroup.getLocationsForJobs()

        self.assertEqual(len(result), 2)
        self.assertEqual("site1" in result, True)
        self.assertEqual("site2" in result, True)

        testJobGroupA = self.createLargerTestJobGroup(commitFlag=True)

        result = testJobGroupA.getLocationsForJobs()

        self.assertEqual(len(result), 2)
        self.assertEqual("site1" in result, True)
        self.assertEqual("site2" in result, True)

        return

    def testGetGroupsByJobStateDAO(self):
        """
        _testGetGroupsByJobStateDAO_

        Verify that the GetGrounsByJobState DAO does what it is supposed to do.
        """
        testWorkflow = Workflow(spec="spec.xml", owner="Simon",
                                name="wf001", task="Test")
        testWorkflow.create()

        testWMBSFileset = WMBSFileset(name="TestFileset")
        testWMBSFileset.create()

        testSubscription = Subscription(fileset=testWMBSFileset,
                                        workflow=testWorkflow)
        testSubscription.create()

        testJobGroupA = JobGroup(subscription=testSubscription)
        testJobGroupA.create()
        testJobGroupB = JobGroup(subscription=testSubscription)
        testJobGroupB.create()

        testJobA = Job(name="TestJobA")
        testJobB = Job(name="TestJobB")

        testJobGroupA.add(testJobA)
        testJobGroupB.add(testJobB)

        testJobGroupA.commit()
        testJobGroupB.commit()

        myThread = threading.currentThread()
        daofactory = DAOFactory(package="WMCore.WMBS",
                                logger=myThread.logger,
                                dbinterface=myThread.dbi)

        stateChangeAction = daofactory(classname="Jobs.ChangeState")
        testJobA["state"] = "complete"
        testJobB["state"] = "executing"
        stateChangeAction.execute(jobs=[testJobA, testJobB])

        jobGroupAction = daofactory(classname="JobGroup.GetGroupsByJobState")
        jobGroups = jobGroupAction.execute(jobState="complete")

        assert len(jobGroups) == 1, \
            "Error: Wrong number of job groups returned."
        assert jobGroups[0] == testJobGroupA.id, \
            "Error: Wrong job group returned."

        return


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
Test WMFactory module
"""






import logging
import os
import threading
import unittest

from WMCore.WMFactory import WMFactory
from WMQuality.TestInit import TestInit

class WMFactoryTest(unittest.TestCase):
    """
    TestCase for Registry module
    """

    _log_level = 'debug'
    # for now set the relative path
    _repository = 'WMCore_t.WMFactory_t'


    def setUp(self):
        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        logging.debug("loading repositories")

    def testRegistry(self):
        logging.debug("Initialize registry")
        # somewhere (e.g. during initialization) you create a registry for queries
        registry = WMFactory('my_test_registry', WMFactoryTest._repository)

        testRegisteredFile = \
        registry.loadObject('TestCollection_1.TestRegistryFile1')
        parameters = {'par1':'val1', 'par2':'val2'}
        result = testRegisteredFile.doSomething(parameters)
        self.assertEqual(parameters, result)
        # this time from cache
        testRegisteredFile = \
        registry.loadObject('TestCollection_1.TestRegistryFile2')
        parameters = {'par3':'val3', 'par4':'val4'}
        result = testRegisteredFile.doSomething(parameters)
        self.assertEqual(parameters, result)

        testRegisteredFile = \
        registry.loadObject('TestCollection_2.TestRegistryFile1')
        parameters = {'par5':'val5', 'par6':'val6'}
        result = testRegisteredFile.doSomething(parameters)
        self.assertEqual(parameters, result)

        testRegisteredFile = \
        registry.loadObject('TestCollection_2.TestRegistryFile2')
        parameters = {'par7':'val7', 'par8':'val8'}
        result = testRegisteredFile.doSomething(parameters)
        self.assertEqual(parameters, result)
        logging.debug("testing cache")
        testRegisteredFile = \
        registry.loadObject('TestCollection_2.TestRegistryFile2')
        testRegisteredFile = \
        registry.loadObject('TestCollection_2.TestRegistryFile2')
        testRegisteredFile = \
        registry.loadObject('TestCollection_2.TestRegistryFile2')
        logging.debug("testing multiple instances (no cache)")
        testRegisteredFileInst1 = \
        registry.loadObject('TestCollection_2.TestRegistryFile2', \
        storeInCache = False)
        testRegisteredFileInst2 = \
        registry.loadObject('TestCollection_2.TestRegistryFile2', \
        storeInCache = False)
        testRegisteredFileInst3 = \
        registry.loadObject('TestCollection_2.TestRegistryFile2', \
        storeInCache = False)

        parameters = {'par9':'val9', 'par10':'val10'}
        result = testRegisteredFileInst1.doSomething(parameters)
        self.assertEqual(parameters, result)
        parameters = {'par11':'val11', 'par12':'val12'}
        result = testRegisteredFileInst2.doSomething(parameters)
        self.assertEqual(parameters, result)
        parameters = {'par13':'val13', 'par14':'val14'}
        result = testRegisteredFileInst3.doSomething(parameters)
        self.assertEqual(parameters, result)

        logging.debug("Retrieving registry from thread attribute and load objects")
        logging.debug("this way all objects in this thread can use the registry")
        myThread = threading.currentThread()
        registry = myThread.factory['my_test_registry']
        parameters = {'par15':'val15', 'par16':'val16'}
        testRegisteredFile = \
        registry.loadObject('TestCollection_3.TestRegistryFile1', \
        args = parameters)
        result = testRegisteredFile.doSomething(parameters)
        self.assertEqual(parameters, result)

if __name__ == '__main__':
    unittest.main()



"""
_DiscreteSummaryHistogram_t_

Unit test module for the DiscreteSummaryHistogram module.
It checks the sanity of the histogram constructions,
corner cases and its compatibility with JSON data published
to couch.

Created on Nov 20, 2012

@author: dballest
"""

from builtins import range
import unittest
import os

from WMCore.DataStructs.MathStructs.DiscreteSummaryHistogram import DiscreteSummaryHistogram
from WMCore.Database.CMSCouch import Database
from WMQuality.TestInitCouchApp import TestInitCouchApp as TestInit

class DiscreteSummaryHistogramTest(unittest.TestCase):

    def setUp(self):
        """
        _setUp_

        Setup a couch database for testing
        of produced JSON
        """

        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setupCouch("histogram_dump_t")
        self.histogramDB = Database(url=os.getenv("COUCHURL") ,dbname = "histogram_dump_t")

    def tearDown(self):
        """
        _tearDown_

        Clean the couch
        """
        self.testInit.tearDownCouch()

    def testA_BasicTest(self):
        """
        _testA_BasicTest_

        Build a histogram from a set of discrete data. Check
        that the statistic properties in the histogram are accurate,
        and that this can become a document an uploaded to couch
        """
        # Try and empty one
        histogram = DiscreteSummaryHistogram('SomeTitle', 'Categories')
        histogramJSON = histogram.toJSON()

        self.assertEqual(histogramJSON["title"], "SomeTitle")
        self.assertEqual(histogramJSON["xLabel"], "Categories")
        self.assertFalse(histogramJSON["continuous"])
        self.assertEqual(len(histogramJSON["data"]), 0)
        self.assertEqual(histogramJSON["average"], {})
        self.assertEqual(histogramJSON["stdDev"], {})

        histogram = DiscreteSummaryHistogram('SomeTitle', 'Categories')

        for _ in range(5):
            histogram.addPoint("CategoryA", "FeatureA")
            histogram.addPoint("CategoryB", "FeatureB")

        for _ in range(17):
            histogram.addPoint("CategoryA", "FeatureB")
            histogram.addPoint("CategoryC", "FeatureB")

        for _ in range(3):
            histogram.addPoint("CategoryC", "FeatureA")

        jsonHistogram = histogram.toJSON()

        # Average/stdDev per feature:
        # FeatureA: avg = 2.7 stdev = 2.05
        # FeatureB: avg = 13 stdev = 5.66
        self.assertAlmostEqual(jsonHistogram["average"]["FeatureA"], 2.7, places = 1)
        self.assertAlmostEqual(jsonHistogram["average"]["FeatureB"], 13, places = 1)
        self.assertAlmostEqual(jsonHistogram["stdDev"]["FeatureA"], 2.05, places = 1)
        self.assertAlmostEqual(jsonHistogram["stdDev"]["FeatureB"], 5.66, places = 1)
        self.assertEqual(jsonHistogram["data"]["CategoryA"]["FeatureA"], 5)
        self.assertEqual(jsonHistogram["data"]["CategoryA"]["FeatureB"], 17)
        self.assertEqual(jsonHistogram["data"]["CategoryB"]["FeatureA"], 0)
        self.assertEqual(jsonHistogram["data"]["CategoryB"]["FeatureB"], 5)
        self.assertEqual(jsonHistogram["data"]["CategoryC"]["FeatureA"], 3)
        self.assertEqual(jsonHistogram["data"]["CategoryC"]["FeatureB"], 17)

        # Test couch
        # Try to commit it to couch
        jsonHistogram["_id"] = jsonHistogram["title"]
        self.histogramDB.commitOne(jsonHistogram)

        storedJSON = self.histogramDB.document("SomeTitle")
        self.assertEqual(len(storedJSON["data"]), 3)

        return

if __name__ == "__main__":
    unittest.main()



import logging
import unittest

from WMQuality.WebTools.RESTBaseUnitTest import RESTBaseUnitTest
from WMQuality.WebTools.RESTClientAPI import methodTest
from WMQuality.WebTools.RESTServerSetup import DefaultConfig


class NestedModelTest(RESTBaseUnitTest):

    def initialize(self):
        self.config = DefaultConfig('WMCore_t.WebTools_t.DummyNestedModel')
        do_debug = True

        self.config.Webtools.environment = 'development'
        if do_debug:
            self.config.Webtools.error_log_level = logging.DEBUG
            self.config.Webtools.access_log_level = logging.DEBUG
        else:
            self.config.Webtools.error_log_level = logging.WARNING
            self.config.Webtools.access_log_level = logging.WARNING

        self.urlbase = self.config.getServerUrl()

    def testOuterFooPass(self):
        verb = 'GET'
        url = self.urlbase + 'foo'
        output = {'code': 200, 'data': b'"foo"'}
        expireTime = 3600
        methodTest(verb, url, output=output, expireTime=expireTime)

        url = self.urlbase + 'foo/test'
        output = {'code': 200, 'data': b'"foo test"'}
        methodTest(verb, url, output=output, expireTime=expireTime)

        url = self.urlbase + 'foo'
        request_input = {'message': "test"}
        output = {'code': 200, 'data': b'"foo test"'}
        methodTest(verb, url, request_input=request_input, output=output, expireTime=expireTime)

    def testInnerPingPass(self):
        verb = 'GET'
        url = self.urlbase + 'foo/ping'
        output = {'code': 200, 'data': b'"ping"'}
        expireTime = 3600

        methodTest(verb, url, output=output, expireTime=expireTime)

    def testOuterFooError(self):
        verb = 'GET'
        url = self.urlbase + 'foo/123/567'
        output = {'code': 400}
        methodTest(verb, url, output=output)

    def testInnerPingError(self):
        verb = 'GET'
        url = self.urlbase + 'foo/123/ping'
        output = {'code': 400}
        methodTest(verb, url, output=output)

        url = self.urlbase + 'foo/ping/123'
        methodTest(verb, url, output=output)


if __name__ == "__main__":
    unittest.main()



"""
ReqMgr unittest configuration file.

"""

# ---------------------------------------------------------------------------
# this entire section copied from:
#     deployment/reqmgr2/config.py


import socket
from os import path, getenv

from WMCore.Configuration import Configuration

HOST = socket.gethostname().lower()
BASE_URL = "@@BASE_URL@@"
DBS_INS = "@@DBS_INS@@"
COUCH_URL = "%s/couchdb" % BASE_URL
LOG_DB_URL = "%s/wmstats_logdb" % COUCH_URL
LOG_REPORTER = "reqmgr2"

ROOTDIR = __file__.rsplit('/', 3)[0]

config = Configuration()

main = config.section_("main")
srv = main.section_("server")
srv.thread_pool = 30
srv.accepted_queue_size = -1
srv.accepted_queue_timeout = 0
main.application = "reqmgr2"
main.port = 9988  # main application port it listens on
main.index = "ui"
# Defaults to allow any CMS authenticated user. Write APIs should require
# additional roles in CRIC (i.e. "Admin" role for the "ReqMgr" group)
main.authz_defaults = {"role": None, "group": None, "site": None}
main.log_screen = True

tools = main.section_("tools")
# provide CherryPy monitoring under: <hostname>/reqmgr2/data/stats
tools.section_("cpstats").on = False
tools.section_("cms_auth").key_file = "%s/auth/wmcore-auth/header-auth-key" % ROOTDIR

# this is where the application will be mounted, where the REST API
# is reachable and this features in CMS web frontend rewrite rules
app = config.section_(main.application)  # string containing "reqmgr2"
app.admin = "cms-service-webtools@cern.ch"
app.description = "CMS data operations Request Manager."
app.title = "CMS Request Manager (ReqMgr)"

views = config.section_("views")

# practical to have this kind of configuration values not in
# service related RPM (difficult/impossible to change in CMS web
# deployment) but in the deployment configuration for the service

# redirector for the REST API implemented handlers
data = views.section_("data")
data.object = "WMCore.ReqMgr.Service.RestApiHub.RestApiHub"
# The couch host is defined during deployment time.
data.couch_host = COUCH_URL
# main ReqMgr CouchDB database containing all requests with spec files attached
data.couch_reqmgr_db = "reqmgr_workload_cache"
# ReqMgr database containing groups, teams, software, etc
data.couch_reqmgr_aux_db = "reqmgr_auxiliary"
# ConfigCache - database with configuration documents
data.couch_config_cache_db = "reqmgr_config_cache"
data.couch_workload_summary_db = "workloadsummary"
data.couch_wmstats_db = "wmstats"
data.couch_acdc_db = "acdcserver"
data.couch_workqueue_db = "workqueue"
data.central_logdb_url = LOG_DB_URL
data.log_reporter = LOG_REPORTER
# Fake permissions for unit tests
data.authorized_roles = {"admin": {'role': ["admin_role"], 'group': ["admin_group"]},
                         "ops": {'role': ["admin_role", "ops_role"],
                                 'group': ["admin_group", "ops_group"]},
                         "ppd": {'role': ["admin_role", "ops_role", "ppd_role"],
                                 'group': ["admin_group", "ops_group", "ppd_group"]}}
data.authz_by_status = [{"permission": "admin",
                         "statuses": ["acquired", "running-open", "running-closed", "completed", "aborted-completed",
                                      "failed", "rejected-archived", "aborted-archived", "normal-archived"]},
                        {"permission": "ops",
                         "statuses": ["assigned", "staging", "staged", "force-complete",
                                      "closed-out", "announced"]},
                        {"permission": "ppd",
                         "statuses": ["new", "assignment-approved", "rejected", "aborted", "NO_STATUS"]}]

# number of past days since when to display requests in the default view
data.default_view_requests_since_num_days = 30  # days
# resource to fetch CMS software versions and scramarch info from
data.tag_collector_url = "https://cmssdt.cern.ch/SDT/cgi-bin/ReleasesXML"
# another source at TC, returns directly JSON, but strangely formatted (e.g.
# keys are not present at easy item but defined in a dedicated item ...)
# https://cmssdt.cern.ch/tc/getReleasesInformation?release_state=Announced

# use dbs testbed for private vm test
data.dbs_url = "https://cmsweb-testbed.cern.ch/dbs/int/global/DBSReader"

# web user interface
ui = views.section_("ui")
ui.object = "WMCore.ReqMgr.WebGui.FrontPage.FrontPage"
ui.static_content_dir = path.join(path.abspath(ROOTDIR),
                                  "apps",
                                  main.application,
                                  "data")

# end of deployment/reqmgr2/config.py
# ---------------------------------------------------------------------------


# ---------------------------------------------------------------------------
# this entire section copied from:
#     deployment/reqmgr2/config-localhost.py


# localhost running additions, no authentication
config.main.section_("tools")
config.main.tools.section_("cms_auth")
config.main.server.socket_host = "127.0.0.1"
config.main.server.environment = "staging"  # must not be "production"
# config.main.tools.cms_auth.policy = "dangerously_insecure"

# go up /deployment/reqmgr2/__file__
first_part = path.abspath(ROOTDIR)
config.views.ui.static_content_dir = path.join(first_part, "WMCore/src/data")

config.views.data.couch_host = getenv("COUCHURL", None)

# end of deployment/reqmgr2/config-localhost.py
# ---------------------------------------------------------------------------
